import numpy as np


# TODO: ensure same ``frate`` explanation for all functions
#       - obtained with SpikeEpochs.spike_rate or SpikeEpochs.spike_density
#       - dimensions: ..., ..., ...
# TODO: adapt for multiple cells
# TODO: make more universal (do not require 'time' and 'trial' dimensions)
def explained_variance(frate, groupby, kind='omega'):
    """Calculate percentage of explained variance (PEV) effect size.

    Parameters
    ----------
    frate : xarray.DataArray
        Firing rate data.
    groupby : str
        Name of the grouping variable. It is assumed that the variable is
        categorical with at least two levels and is specified as a coordinate
        for the trial dimension.
    kind : str
        Type of effect size to calculate. Can be either 'omega' or 'eta' for
        omega squared and eta squared respectively.

    Returns
    -------
    es : xarray.DataArray
        Explained variance effect size.
    """
    assert kind in ('omega', 'eta'), 'kind must be either "omega" or "eta"'
    is_omega = kind == 'omega'

    global_avg = frate.mean(dim='trial')
    has_time = 'time' in frate.dims
    if has_time:
        n_times = len(frate.coords['time'])

    groups, per_group = np.unique(frate.coords[groupby], return_counts=True)
    n_groups = len(groups)

    SS_total = ((frate - global_avg) ** 2).sum(dim='trial')
    SS_between = (np.zeros((n_groups, n_times))
                  if has_time else np.zeros(n_groups))

    if is_omega:
        SS_within = SS_between.copy()

    for idx, (label, arr) in enumerate(frate.groupby(groupby)):
        # are group labels always sorted when using .groupby?
        group_avg = arr.mean(dim='trial')
        SS_between[idx] = per_group[idx] * (group_avg - global_avg) ** 2

        if is_omega:
            within_group = ((arr - group_avg) ** 2).sum(dim='trial')
            SS_within[idx] = within_group

    SS_between = SS_between.sum(axis=0)

    if not is_omega:
        es = SS_between / SS_total
        es.name = 'eta squared'
    else:
        df = n_groups - 1
        n_trials = len(frate.coords['trial'])
        MSE = SS_within.sum(axis=0) / n_trials
        es = (SS_between - df * MSE) / (SS_total + MSE)
        es.name = 'omega squared'

    return es


def depth_of_selectivity(frate, groupby, ignore_below=1e-15):
    '''Compute depth of selectivity for given category.

    Parameters
    ----------
    frate : xarray
        Xarray with firing rate data.
    groupby : str
        Name of the dimension to group by and calculate depth of selectivity
        for.
    ignore_below : float
        Ignore values below this threshold. This is useful when spike density
        is passed in ``frate`` - due to numerical errors, some values may be
        very small or even negative but not exactly zero. Such values can lead
        to depth of selectivity being far greater than 1. Default is ``1e-15``.

    Returns
    -------
    selectivity : xarray
        Xarray with depth of selectivity.
    '''
    if ignore_below > 0:
        frate = frate.copy()
        msk = frate.values < ignore_below
        frate.values[msk] = 0.

    avg_by_probe = frate.groupby(groupby).mean(dim='trial')
    n_categories = len(avg_by_probe.coords[groupby])
    r_max = avg_by_probe.max(dim=groupby)

    singleton = r_max.shape == ()
    if singleton and r_max.item() == 0:
        return 0, avg_by_probe

    numerator = n_categories - (avg_by_probe / r_max).sum(dim=groupby)
    selectivity = numerator / (n_categories - 1)
    selectivity.name = 'depth of selectivity'

    if not singleton:
        bad_selectivity = r_max < ignore_below
        if (bad_selectivity).any():
            selectivity[bad_selectivity] = 0

    return selectivity, avg_by_probe


def compute_selectivity_continuous(frate, compare='image', n_perm=500,
                                   n_jobs=1, min_Hz=False):
    '''
    Compute population selectivity for specific experimental category.

    Uses permutation test to compute selectivity p-values.

    Parameters
    ----------
    frate : xarray.DataArray
        Firing rate of the neurons.
    compare : str
        Metadata category to compare. Defaults to ``'image'``.
    n_perm : int
        Number of permutations to use for permutation test. Defaults to
        ``500``.
    n_jobs : int
        Number of parallel jobs. No parallel computation is done when
        ``n_jobs=1`` (default).
    min_Hz : 0.1 | bool
        Minimum spiking rate threshold (in Hz). Cells below this threshold will
        be ignored.

    Returns
    -------
    selectivity : dict of xarray.DataArray
        Dictionary of DataArrays with selectivity for each cell. The following
        keys are used:
        * ``'stat'`` - selectivity statistic (t values), DataArray with
          dimensions ``('cell', 'time')`` (unless time was not present in the
          ``frate``)
        * ``'thresh'`` - 95% significance thresholds from permutation test:
          lower, negative (2.5%) and higher, positive (97.5%) tails. DataArray
          with dimensions ``('tail', 'cell', 'time')`` (unless time was not
          present in the ``frate``)
        * ``'perm'`` - selectivity statistic for each permutation. DataArray
          with dimensions ``('perm', 'cell', 'time')`` (unless time was not
          present in the ``frate``)
    '''
    import xarray as xr
    from .stats import permutation_test

    has_time = 'time' in frate.dims

    # select cells
    if min_Hz:
        reduce_dims = 'trial' if not has_time else ('trial', 'time')
        msk = frate.mean(dim=reduce_dims) >= min_Hz
        frate = frate.sel(cell=msk)

    frate_dims = ['trial', 'cell']
    if has_time:
        frate_dims.append('time')
    frate = frate.transpose(*frate_dims)

    # permutations
    # ------------
    arrs = [arr for _, arr in frate.groupby(compare)]
    results = permutation_test(
        *arrs, paired=False, n_perm=n_perm,
        return_pvalue=False, return_distribution=True, n_jobs=n_jobs)

    # turn to xarray
    # --------------
    cells = frate.cell.values
    if 'subject' in frate.attrs:
        subj = frate.attrs['subject']
        cells = ['_'.join([subj, x]) for x in cells]

    # perm
    dims = ['perm', 'cell']
    coords = {'perm': np.arange(n_perm) + 1,
              'cell': cells}
    if has_time:
        dims.append('time')
        coords['time'] = frate.time.values

    results['dist'] = xr.DataArray(data=results['dist'], dims=dims,
                                   coords=coords, name='t value')

    # stat
    coords = {k: coords[k] for k in dims[1:]}
    results['stat'] = xr.DataArray(
        data=results['stat'], dims=dims[1:], coords=coords, name='t value')

    # thresh
    results['thresh'] = np.stack(results['thresh'], axis=0)
    dims2 = ['tail'] + dims[1:]
    coords.update({'tail': ['pos', 'neg']})
    results['thresh'] = xr.DataArray(
        data=results['thresh'], dims=dims2, coords=coords, name='t value')

    # add cell coords
    for key in results.keys():
        copy_coords = ['region', 'region2', 'anat', 'channel', 'cluster']
        copy_coords = [coord for coord in copy_coords if coord in frate.coords]
        coords = {coord: ('cell', frate.coords[coord].values)
                  for coord in copy_coords}
        results[key] = results[key].assign_coords(coords)

    return results


# TODO: add njobs
# TODO: refactor to separate cluster-based and cell selection
# TODO: create more progress bars and pass to cluster_based_test
# TODO: use calculate_dos to ignore DoS calculation
def cluster_based_selectivity(frate, compare, spk=None,
                              cluster_entry_pval=0.05, n_permutations=1000,
                              n_stat_permutations=0, pbar='notebook',
                              correct_window=0., min_cluster_pval=0.1,
                              calculate_pev=False, calculate_dos=True,
                              calculate_peak_pev=False, stat_fun=None,
                              baseline_window=(-0.5, 0)):
    '''Calculate category-sensitivity of neurons using cluster-based tests.

    Performs cluster-based ANOVA on firing rate of each neuron to test
    category-selectivity of the neurons.

    Parameters
    ----------
    frate : xarray.DataArray
        Xarray with spike rate  or spike density containing ``'cell'``,
        ``'trial'`` and ``'time'`` dimensions.
    compare : str
        Dimension labels specified for ``'trial'`` dimension that constitutes
        categories to test selectivity for.
    spk : SpikeEpochs
        Spikes object to use when calculating depth of sensitivity (DoS) or
        percentage of explained variance (PEV) in cluster time window.
        Not used if ``calculate_pev`` and ``calculate_dos`` are ``False``.
    cluster_entry_pval : float
        P value used as a cluster-entry threshold. The default is ``0.05``.
    n_permutations : int
        Number of permutations to use for permutation test. The default is
        ``1000``.
    n_stat_permutations : int
        Number of permutations to use for non-parametric calculation of the p
        value for the test statistic (see ``stat_fun``). The default is ``0``,
        which means that the p value is calculated parametrically. Using
        non-parametric calculation is slower but more resistant to parametric
    pbar : str | tqdm progressbar
        Progressbar to use. The default is ``'notebook'`` which creates a
        ``tqdm.notebook.tqdm`` progressbar.
    correct_window : float
        The cluster time window will be extended by this value when calculating
        window average statistics (DoS, PEV). The default is ``0.``.
    min_cluster_pval : float
        Minimum p value for a cluster to be stored in the results dataframe.

    Returns
    -------
    df_cluster : pandas.DataFrame
        Dataframe with category-sensitivity information.
    '''
    import pandas as pd
    import xarray as xr

    from .stats import cluster_based_test
    from .viz import check_modify_progressbar

    n_cells = len(frate)
    corrwin = [-correct_window, correct_window]
    pbar = check_modify_progressbar(pbar, total=n_cells)

    # init dataframe for storing results
    cols = ['pval', 'window', 'preferred', 'n_preferred', 'FR_preferred']
    cols = ['neuron', 'region', 'region_short', 'anat', 'cluster'] + cols
    if calculate_pev:
        cols += ['pev']
    if calculate_dos:
        cols += ['DoS']
    if calculate_peak_pev:
        cols += ['peak_pev']
    if baseline_window is not None:
        cols += ['FR_vs_baseline']

    df_cluster = pd.DataFrame(columns=cols)

    for pick, fr_cell in enumerate(frate):
        cell_name = fr_cell.cell.item()
        row_idx = df_cluster.shape[0]

        # TODO - pass relevant arrays
        _, clusters, pval = cluster_based_test(
            fr_cell, compare=compare, cluster_entry_pval=cluster_entry_pval,
            paired=False, stat_fun=stat_fun, n_permutations=n_permutations,
            n_stat_permutations=n_stat_permutations, progress=False)

        # process clusters
        # TODO - separate function
        n_clusters = len(pval)
        if n_clusters > 0:

            # select clusters to save in df
            good_clst = pval <= min_cluster_pval
            clst_idx = np.where(good_clst)[0]

            for ord_idx, idx in enumerate(clst_idx):
                # get cluster time window
                clst_msk = clusters[idx]
                this_pval = pval[idx]
                twin = fr_cell.time[clst_msk].values[[0, -1]]
                twin_str = '{:.03f} - {:.03f}'.format(*twin)

                # get anatomical information
                if 'region' in fr_cell.coords:
                    region_name = fr_cell.region.item()

                    if 'region2' in fr_cell.coords:
                        short_name = fr_cell.region2.item()
                    else:
                        short_name = region_name[1:-1]
                else:
                    region_name = None
                    short_name = None

                if 'anat' in fr_cell.coords:
                    anat_name = fr_cell.anat.item()
                else:
                    anat_name = None

                df_idx = row_idx + ord_idx
                df_cluster.loc[df_idx, 'neuron'] = cell_name
                df_cluster.loc[df_idx, 'region'] = region_name
                df_cluster.loc[df_idx, 'region_short'] = short_name
                df_cluster.loc[df_idx, 'anat'] = anat_name
                df_cluster.loc[df_idx, 'cluster'] = ord_idx + 1
                df_cluster.loc[df_idx, 'pval'] = this_pval
                df_cluster.loc[df_idx, 'window'] = twin_str

                # calculate depth of selectivity for each selective window
                tmin, tmax = twin + corrwin
                if spk is not None:
                    frate_avg = spk.spike_rate(
                        picks=pick, step=False, tmin=tmin, tmax=tmax
                        ).isel(cell=0)

                    # TODO: for GammBur I used only correct trials,
                    #       but it may not always make sense...
                    # .sel(trial=frate_avg.ifcorrect)

                    if baseline_window is not None:
                        frate_bsln = spk.spike_rate(
                            picks=pick, step=False, tmin=baseline_window[0],
                            tmax=baseline_window[1]).isel(cell=0)

                else:
                    frate_avg = fr_cell.sel(time=slice(tmin, tmax)).mean(
                        dim='time')  # .sel(trial=frate.ifcorrect)

                    if baseline_window is not None:
                        frate_bsln = fr_cell.sel(
                            time=slice(baseline_window[0], baseline_window[1])
                            ).mean(dim='time')

                depth, preferred = depth_of_selectivity(
                    frate_avg, groupby=compare)
                if isinstance(depth, xr.DataArray):
                    depth = depth.item()

                # TODO - this also could be moved outside
                # find preferred categories
                perc_pref = preferred / preferred.max()
                pref_idx = np.where(perc_pref >= 0.75)[0]
                perc_pref_sel = perc_pref.values[pref_idx].argsort()[::-1]

                # TODO - use category labels in preferred!
                pref = pref_idx[perc_pref_sel].tolist()
                pref_str = str(pref)

                # calculate PEV and peak PEV
                if calculate_pev:
                    pev = explained_variance(
                        frate_avg, compare).item()

                if calculate_peak_pev:
                    peak_pev = explained_variance(
                        fr_cell.sel(time=slice(tmin, tmax)),
                        compare
                        ).max().item()

                if baseline_window is not None:
                    FR_vs_bsln = (preferred.max().item()
                                  / frate_bsln.mean().item())

                df_cluster.loc[df_idx, 'preferred'] = pref_str
                df_cluster.loc[df_idx, 'n_preferred'] = len(pref)
                df_cluster.loc[df_idx, 'FR_preferred'] = preferred.max().item()
                df_cluster.loc[df_idx, 'DoS'] = depth

                if calculate_pev:
                    df_cluster.loc[df_idx, 'PEV'] = pev

                if calculate_peak_pev:
                    df_cluster.loc[df_idx, 'peak_PEV'] = peak_pev

                if baseline_window is not None:
                    df_cluster.loc[df_idx, 'FR_vs_baseline'] = FR_vs_bsln

        pbar.update(1)

    # make sure that the dataframe columns are of the right type
    df_cluster = df_cluster.infer_objects()
    return df_cluster


def assess_selectivity(df_cluster, min_cluster_p=0.05,
                       window_of_interest=(0.1, 1), min_time_in_window=0.2,
                       min_depth_of_selectivity=None, min_pev=None,
                       min_peak_pev=None, min_FR_vs_baseline=None,
                       min_FR_preferred=None):
    pval_msk = df_cluster.pval <= min_cluster_p

    def eval_column(column_name, min_value):
        msk = (np.ones_like(pval_msk) if min_value is None
               else df_cluster[column_name] >= min_value)
        return msk

    if min_time_in_window is not None:
        df_cluster = compute_time_in_window(df_cluster, window_of_interest)
        in_toi_msk = df_cluster.in_toi >= min_time_in_window
    else:
        in_toi_msk = np.ones_like(pval_msk)

    dos_msk = eval_column('DoS', min_depth_of_selectivity)
    pev_msk = eval_column('PEV', min_pev)
    peak_pev_msk = eval_column('peak_PEV', min_peak_pev)
    fr_vs_bsln_msk = eval_column('FR_vs_baseline', min_FR_vs_baseline)
    fr_pref_msk = eval_column('FR_preferred', min_FR_preferred)

    # check if cluster time window can be deemed as selective
    is_sel = (pval_msk & pev_msk & in_toi_msk & dos_msk & peak_pev_msk
              & fr_vs_bsln_msk & fr_pref_msk)

    df_cluster['selective'] = is_sel
    return df_cluster


def _compute_time_in_window(clst_window, window_of_interest):
    # clst_limits = times[clst_msk].values[[0, -1]]
    twin = [0., 0.]
    clst_limits = _parse_window(clst_window)
    twin[0] = min(max(window_of_interest[0], clst_limits[0]),
                      window_of_interest[1])
    twin[1] = max(min(window_of_interest[1], clst_limits[1]),
                      window_of_interest[0])
    return twin[1] - twin[0]


def _parse_window(window_str):
    start, fin = [float(x) for x in window_str.split(' - ')]
    return start, fin


def compute_time_in_window(df_cluster, window_of_interest):
    '''
    Compute time spent in the window of interest for each cluster.

    Parameters
    ----------
    df_cluster : pandas.DataFrame
        Dataframe with cluster information.
    window_of_interest : tuple of float
        Time window of interest.

    Returns
    -------
    df_cluster : pandas.DataFrame
        Dataframe with cluster information, including time spent in the window
        of interest.
    '''
    df_cluster.loc[:, 'in_toi'] = df_cluster.window.apply(
        _compute_time_in_window, args=(window_of_interest,))
    return df_cluster


# TODO: compare with time resolved selectivity
def compute_selectivity_windows(spk, windows=None, compare='image',
                                baseline=None, test='kruskal', n_perm=2000,
                                progress=True):
    '''
    Compute selectivity for each cell in specific time windows.

    Parameters
    ----------
    spk : SpikeEpochs
        Spike epochs object.
    windows : dict of tuples
        Dictionary with keys being the names of the windows and values being
        ``(start, end)`` tuples of window time limits in seconds.
    compare : str
        Metadata category to compare. Defaults to ``'image'``. If ``None``,
        the average firing rate is compared to the baseline.
    baseline : None | xarray.DataArray
        Baseline firing rate to compare to.
    test : str
        Test to use for computing selectivity. Can be ``'kruskal'`` or
        ``'permut'``. Defaults to ``'kruskal'``.
    n_perm : int
        Number of permutations to use for permutation test.
    progress : bool | str
        Whether to show a progress bar. If string, it can be ``'text'`` for
        text progress bar or ``'notebook'`` for a notebook progress bar.

    Returns
    -------
    selectivity : dict of pandas.DataFrame
        Dictionary of dataframes with selectivity for each cell. Each
        dictionary key and each dataframe corresponds to a time window of given
        name.
    '''
    import pandas as pd
    from scipy.stats import kruskal
    from sarna.utils import progressbar

    use_test = kruskal if test == 'kruskal' else permutation_test
    test_args = {'n_perm': n_perm} if test == 'permut' else {}

    if windows is None:
        windows = {'early': (0.1, 0.6), 'late': (0.6, 1.1)}

    columns = ['neuron', 'region', f'{test}_stat', f'{test}_pvalue', 'DoS']
    has_region = 'region' in spk.cellinfo
    if not has_region:
        columns.pop(1)

    if compare is not None:
        level_labels = np.unique(spk.metadata[compare])
        n_levels = len(level_labels)
        level_cols = [f'FR_{compare}{lb}' for lb in level_labels]
        level_cols_norm = (list() if baseline is None else
                        [f'nFR_{compare}{lb}' for lb in level_labels])
        columns += level_cols + level_cols_norm
    else:
        columns += ['FR_baseline', 'FR_condition', 'nFR_condition']

    frate, df = dict(), dict()
    for name, limits in windows.items():
        frate[name] = spk.spike_rate(tmin=limits[0], tmax=limits[1],
                                     step=False)
        df[name] = pd.DataFrame(columns=columns)

    n_cells = len(spk)
    n_windows = len(windows.keys())
    pbar = progressbar(progress, total=n_cells * n_windows)
    for cell_idx in range(n_cells):
        if has_region:
            brain_region = spk.cellinfo.loc[cell_idx, 'region']

        for window in windows.keys():
            if has_region:
                df[window].loc[cell_idx, 'region'] = brain_region
            df[window].loc[cell_idx, 'neuron'] = spk.cell_names[cell_idx]

            if not (frate[window][cell_idx].values > 0).any():
                stat, pvalue = np.nan, 1.
            elif compare is not None:
                data = [arr.values for label, arr in
                        frate[window][cell_idx].groupby(compare)]
                stat, pvalue = use_test(*data, **test_args)
            elif compare is None and baseline is not None:
                data = [frate[window][cell_idx].values,
                        baseline[cell_idx].values]
                stat, pvalue = use_test(*data, **test_args)

            df[window].loc[cell_idx, f'{test}_stat'] = stat
            df[window].loc[cell_idx, f'{test}_pvalue'] = pvalue

            # compute DoS
            if compare is not None:
                dos, avg = depth_of_selectivity(frate[window][cell_idx],
                                                by=compare)
                preferred = int(avg.argmax(dim=compare).item())
                df[window].loc[cell_idx, 'DoS'] = dos.item()
                df[window].loc[cell_idx, 'preferred'] = preferred

                # other preferred?
                perc_of_max = avg / avg[preferred]
                other_pref = ((perc_of_max < 1.) & (perc_of_max >= 0.75))
                if other_pref.any().item():
                    txt = ','.join([str(x) for x in np.where(other_pref)[0]])
                else:
                    txt = ''
                df[window].loc[cell_idx, 'preferred_second'] = txt

            # save firing rate
            if baseline is not None:
                base_fr = baseline[cell_idx].mean(dim='trial').item()
                if base_fr == 0:
                    if compare is not None and (avg > 0).any().item():
                        base_fr = avg[avg > 0].min().item()
                    else:
                        base_fr = 1.

            if compare is not None:
                for idx in range(n_levels):
                    fr = avg[idx].item()
                    level = avg.coords[compare][idx].item()
                    df[window].loc[cell_idx, f'FR_{compare}{level}'] = fr

                    if baseline is not None:
                        nfr = fr / base_fr
                        df[window].loc[cell_idx, f'nFR_{compare}{level}'] = nfr
            else:
                average_fr = frate[window][cell_idx].mean(dim='trial').item()
                df[window].loc[cell_idx, 'FR_baseline'] = base_fr
                df[window].loc[cell_idx, 'FR_condition'] = average_fr
                df[window].loc[cell_idx, 'nFR_condition'] = average_fr / base_fr
            pbar.update(1)
    pbar.close()

    for window in windows.keys():
        df[window] = df[window].infer_objects()
        if compare is not None:
            # TODO: not sure if this is the right way to do in all cases
            df[window].loc[:, 'preferred'] = (
                df[window]['preferred'].astype('int'))

    return df, frate



# -*- coding: utf-8 -*-
"""
Scirpt that matches units identified in mixed alignment to corresponding
units in min/max alignment to identify merge candidates.

@author: mmagnuski
"""

# SETTINGS
# --------

# the directory with sorting results - that is after manual curation and
# export using updateSORTINGresults_mm matlab function located in
# psy_screenning-\helpers\sorting_utils)
results_dir = (r'G:\.shortcut-targets-by-id\1XlCWYRlHP0YDbmo3p1NGIC6lN9XZ8l1O\switchorder\derivatives\sorting\sub-W02\ses-main\sub-W02_ses-main_task-switchorder_run-01_sorter-osort_norm-False')

# the directory with raw sorting data - should contain min and max
# subdirectories with respective threshold subdirectories
sorting_dir = (r'C:\Users\mmagnuski\Dropbox\PROJ\Labianka\switchorder\sorting\sub-W02\osorted')
# the sorting threshold used (defines subdirectory in min and max directories)
sorting_threshold = '5'

# whether to save figures (use only for testing purposes, many figures will
#                          be created so the script may bet much slower)
save_figures = True

# where to save figures
fig_dir = (r'C:\Users\mmagnuski\Dropbox\PROJ\Labianka\sorting\merge_'
           'candidates_sub-W02')

# where to save results
save_spk_fname = 'all_channels.mat'
save_spk_dir = (r'C:\Users\mmagnuski\Dropbox\PROJ\Labianka\sorting'
                '\sub-W02_switchorder_after_cleanup')

# neg / pos peak ratio above which (or below its inverse) a waveform is
# classified as neg / pos polarity. This detected polarity is 
polarity_threshold = 1.333

# coincidence_threshold defines above what spike coincidence value
# a matching unit in max/min alignment will be accepted
# the default is 50% (0.5) - and seems to be a good choice
coincidence_threshold = 0.5


# %%
import os.path as op
from warnings import warn

import numpy as np
import matplotlib.pyplot as plt

import pylabianca as pln

# read spike times and waveforms
spk = pln.io.read_osort(results_dir, waveform=True)


def plot_waveform_polarity_choice(spk, idx, unit_type):
    fig, ax = plt.subplots(ncols=2, figsize=(10, 5), sharey=True)

    spk.plot_waveform(idx, ax=ax[0], labels=False)
    avg_waveform = spk.waveform[idx].mean(axis=0)
    ax[1].plot(avg_waveform)

    fig.suptitle(f'Classified as {unit_type[idx]}', fontsize=14)
    return fig


# TODO:
# for merge candidate plotting we need:
# mix_spk, match_spk, indices from first, one index from second
# and coincidence for each of the candidates
def plot_merge_candidates(spk, merge_info):

        merge_names = merge_info['cell_names']
        n_merge = len(merge_names)
        spk = spk.copy().pick_cells(merge_names)

        matching_alignment = merge_info['matching_alg']

        fig, ax = plt.subplots(
            ncols=n_merge + 1, figsize=(3 * (n_merge + 1), 5), sharey=True,
            gridspec_kw={'top': 0.8, 'right': 0.98, 'left': 0.06}
        )

        for idx in range(n_merge):
            spk.plot_waveform(idx, ax=ax[idx])

            this_n = spk.waveform[idx].shape[0]
            this_coincid = merge_info['coincidence'][idx]
            ax1_title = (f'cluster id: {spk.cellinfo.loc[idx, "cluster"]} '
                         f'\n(N = {this_n}) coinc: {this_coincid:.2f}')
            ax[idx].set_title(ax1_title, fontsize=12)

        # plot match
        spk_match = merge_info['matching_spk']
        spk_match.plot_waveform(0, ax=ax[-1])
        n_match = spk_match.waveform[0].shape[0]
        ax2_title = (f'Matching unit from {matching_alignment}\n'
                     f'(N = {n_match}) cluster id: {spk_match.cellinfo.loc[0, "cluster"]}')
        ax[-1].set_title(ax2_title, fontsize=12)

        title = (f'Channel {spk.cellinfo.loc[0, "channel"]}\n')
        fig.suptitle(title, fontsize=14)
        return fig


# find mixed alignment channels and test whether every channel has only one
# alignment type
msk_mixed = spk.cellinfo['alignment'] == 'mixed'
mixed_chans = spk.cellinfo.loc[msk_mixed, 'channel'].unique()

# make sure each channel contains only one alignment
chans = spk.cellinfo.channel.unique()
for chan in chans:
    info_sel = spk.cellinfo.query(f'channel == "{chan}"')

    alignments = info_sel.alignment.unique()
    n_alignments = len(alignments)

    if n_alignments > 1:
        msg = f'Channel {chan} contains {n_alignments} alignments:'
        for algn in alignments:
            n_units = (info_sel.alignment == algn).sum()
            msg += f'\n{algn}: {n_units} units; '
        warn(msg)


# main part of the script
# -----------------------
merges = list()

for chan in mixed_chans:
    simil = dict()
    spk_alg = dict()

    info_sel = spk.cellinfo.query(f'channel == "{chan}"')
    cell_idx = info_sel.index.to_numpy()
    cell_names = spk.cell_names[cell_idx]
    spk_sel = spk.copy().pick_cells(cell_idx)

    for check_type in ['pos', 'neg']:

        use_dir = 'max' if check_type == 'pos' else 'min'
        this_sorting_dir = op.join(sorting_dir, use_dir, sorting_threshold)
        spk2 = pln.io.read_osort(
            this_sorting_dir, channels=chan,
            format='standard', progress=False
        )

        # compare coincidence- spk_sel vs spk2
        simil[use_dir] = pln.spike_distance.compute_spike_coincidence_matrix(
            spk_sel, spk2=spk2, tol=0.002, progress=False)
        spk_alg[use_dir] = spk2.copy()

    all_combine = list()
    all_simil = list()
    all_alg = list()
    matches = list()

    for alg in ['min', 'max']:
        msk = simil[alg] > coincidence_threshold
        n_merged = msk.sum(axis=0)
        merge_candidates = n_merged > 1
        if merge_candidates.any():
            combine = msk[:, merge_candidates]
            all_combine.append(combine)
            all_simil.append(simil[alg][:, merge_candidates])
            all_alg.extend([alg] * sum(merge_candidates))
            matches.extend(np.where(merge_candidates)[0].tolist())

    n_comb = len(all_combine)
    if n_comb > 1:
        all_combine = np.concatenate(all_combine, axis=1)
        all_simil = np.concatenate(all_simil, axis=1)
    elif n_comb == 1:
        all_combine = all_combine[0]
        all_simil = all_simil[0]

    if n_comb > 0:
        # what to do if there are overlaps?
        has_overlaps = all_combine.sum(axis=1) > 1
        overlapping_idx = np.where(has_overlaps)[0]
        is_solved = np.zeros(all_combine.shape[1], dtype=bool)
        remove = np.zeros(all_combine.shape[1], dtype=bool)

        for ix in range(len(overlapping_idx)):
            col_mask = all_combine[overlapping_idx[ix], :]

            if not (is_solved[col_mask]).all():
                col_mask = col_mask & ~remove
                if col_mask.sum() < 2:
                    continue

                pair = all_combine[:, col_mask]
                assert pair.shape[1] == 2
                if (pair[:, 0] == pair[:, 1]).all():
                    # check if overlapping merges are identical
                    is_solved[col_mask] = True
                    rem_idx = np.where(col_mask)[0][-1]
                    remove[rem_idx] = True
                else:
                    # else - pick the one with highest average coincidence
                    pair_simil = all_simil[:, col_mask]
                    print('two different suggestions')
                    print(pair_simil)

                    pair_simil[~pair] = np.nan
                    avg_simil = np.nanmean(pair_simil, axis=0)
                    is_solved[col_mask] = True
                    rem_idx = np.where(col_mask)[0][avg_simil.argmin()]
                    remove[rem_idx] = True

        all_combine = all_combine[:, ~remove]
        all_simil = all_simil[:, ~remove]
        all_alg = np.array(all_alg)[~remove]
        matches = np.array(matches)[~remove]

        # turn combine matrices into cell merge indices
        for idx in range(all_combine.shape[1]):
            msk = all_combine[:, idx]
            merge_info = dict()
            merge_info['channel'] = chan
            merge_info['cell_names'] = cell_names[msk]
            merge_info['cell_idx'] = cell_idx[msk]
            merge_info['coincidence'] = all_simil[:, idx][msk]

            matching_alg = all_alg[idx]
            matching_idx = matches[idx]
            matching_spk = spk_alg[matching_alg].copy().pick_cells(matching_idx)
            merge_info['matching_alg'] = matching_alg
            merge_info['matching_spk'] = matching_spk

            merges.append(merge_info)

            # plot
            if save_figures:
                fname = f'merge_candidates_chan-{chan}_{idx:02g}.png'
                fig = plot_merge_candidates(spk, merge_info)
                plt.savefig(op.join(fig_dir, fname), dpi=100)
                plt.close(fig)



import itertools
import numpy as np

from . import utils, viz
from .utils import (_deal_with_picks, _turn_spike_rate_to_xarray,
                    _symmetric_window_samples, _gauss_kernel_samples,
                    spike_centered_windows)


# TODO: ``tol=None`` could return the distances without thresholding
# TODO: consider other name like ``find_spike_distances``
# TODO: test for presence of numba and use the numba implementation
#       (could also use ``backend`` argument)
# TODO: multiprocessing could be useful
# TODO: the implementation and the API are suboptimal
# TODO: remove the numpy backend?
def compare_spike_times(spk, cell_idx1, cell_idx2, spk2=None, backend='numba',
                        tol=None):
    '''Test concurrence of spike times for Spikes or SpikeEpochs.

    Parameters
    ----------
    spk : SpikeEpochs
        SpikeEpochs object.
    cell_idx1 : int
        Index of the first cell to compare. All spikes of this cell will be
        checked.
    cell_idx2 : int
        Index of the second cell to compare. Spikes of the first cell will
        be tested for concurrence with spikes coming from this cell.
    backend : str
        Backend to use for the computation. Currently only 'numba' is
        supported.
    tol : float
        Concurrence tolerance in seconds. Spikes no further that this will be
        deemed co-occurring. Default is ``None``, which means no concurrence
        threshold will be applied and a distance matrix will be returned.

    Returns
    -------
    distance: float | array of float
        If ``tol`` is ``None`` an array of time distances between closest
        spikes of the two cells is returned. More precisely, for each spike
        from cell 1 a distance to closest cell 2 spike is given. If ``tol`` is
        not ``None`` then a percentage of spikes from first cell that concur
        with spikes from the second cell is given.
    '''
    from .spikes import SpikeEpochs, Spikes

    if isinstance(spk, SpikeEpochs):
        raise NotImplementedError('Sorry compare_spike_times does not work '
                                  'with SpikeEpochs yet.')
    elif isinstance(spk, Spikes):
        if backend == 'numba':
            from ._numba import numba_compare_times
            distances = numba_compare_times(spk, cell_idx1, cell_idx2,
                                            spk2=spk2)
            if tol is not None:
                distances = (distances < tol).mean()
            return distances
        else:
            if spk2 is not None:
                raise NotImplementedError(
                    'Sorry compare_spike_times does not support spk2 '
                    'in the numpy backend.'
                )

            # TODO: rework this
            tms1 = spk.timestamps[cell_idx1] / spk.sfreq
            tms2 = spk.timestamps[cell_idx2] / spk.sfreq
            time_diffs = np.abs(tms1[:, None] - tms2[None, :])
            closest_time1 = time_diffs.min(axis=1)
            return (closest_time1 < tol).mean()


def numpy_compare_times(spk, cell_idx1, cell_idx2):
    tms1 = spk.timestamps[cell_idx1] / spk.sfreq
    tms2 = spk.timestamps[cell_idx2] / spk.sfreq
    time_diffs = np.abs(tms1[:, None] - tms2[None, :])
    closest_time1 = time_diffs.min(axis=1)
    return closest_time1


def compute_spike_coincidence_matrix(spk, spk2=None, tol=0.002, progress=True):
    if progress:
        from tqdm import tqdm

    n_cells = len(spk)
    n_cells2 = n_cells if spk2 is None else len(spk2)

    similarity = np.zeros((n_cells, n_cells2))
    iter_over = tqdm(range(n_cells)) if progress else range(n_cells)

    for cell1 in iter_over:
        for cell2 in range(n_cells2):
            if spk2 is None and cell1 == cell2:
                continue

            simil = compare_spike_times(spk, cell1, cell2, spk2=spk2, tol=tol)
            similarity[cell1, cell2] = simil

    return similarity


def _construct_bins(sfreq, max_lag, bins=None):
    '''Construct bins for cross-correlation histogram.

    Parameters
    ----------
    sfreq : float
        Sampling frequency of the bins. The bin width will be ``1 / sfreq``
        seconds. Used only when ``bins is None``.
    max_lag : float
        Maximum lag in seconds. Used only when ``bins is None``.
    bins : numpy array | None
        Array representing edges of the histogram bins. If ``None`` (default)
        the bins are constructed based on ``sfreq`` and ``max_lag``.

    Returns
    -------
    bins : numpy array
        Array representing edges of the histogram bins.
    '''
    if bins is not None:
        return bins
    else:
        smp_time = 1 / sfreq
        bins = np.arange(-max_lag, max_lag + smp_time / 2, step=smp_time)

        # check if middle bin is zero-centered
        bin_width = np.diff(bins[:2])[0]
        mid_bin = np.abs(bins).argmin()
        diff_to_centered = np.abs(bins[mid_bin] - (bin_width / 2))

        if diff_to_centered > (bin_width / 10):
            bins = bins - diff_to_centered
            bins = np.concatenate([bins, [bins[-1] + bin_width]])

    return bins


# TODO: add option to shuffle trials?
# TODO: add option to shift trials?
def _xcorr_hist_trials(spk, cell_idx1, cell_idx2, sfreq=500., max_lag=0.2,
                       bins=None):
    '''
    Compute cross-correlation histogram between two cells of SpikeEpochs.

    Parameters
    ----------
    spk : pylabianca.spikes.SpikeEpochs
        SpikeEpochs object to use.
    cell_idx1 : ...
        List of cell indices or names to perform cross- and auto- correlations
        for. All combinations of cells will be used.
    cell_idx2 : ...
    sfreq : float
        Sampling frequency of the bins. The bin width will be ``1 / sfreq``
        seconds. Used only when ``bins is None``. Defaults to ``500.``.
    max_lag : float
        Maximum lag in seconds. Used only when ``bins is None``. Defaults
        to ``0.2``.
    bins : numpy array | None
        Array representing edges of the histogram bins. If ``None`` (default)
        the bins are constructed based on ``sfreq`` and ``max_lag``.

    Returns
    -------
    bins : numpy array
        Array representing edges of the histogram bins.
    '''
    from .utils import _get_trial_boundaries

    bins = _construct_bins(sfreq, max_lag, bins=bins)
    max_tri = max([tri.max() for tri in spk.trial])
    xcorr = np.zeros((max_tri + 1, len(bins) - 1), dtype=int)

    # TODO: this could be done once at the beginning
    trial_boundaries1, tri_num1 = _get_trial_boundaries(spk, cell_idx1)
    if cell_idx1 == cell_idx2:
        autocorr = True
        trial_boundaries2, tri_num2 = trial_boundaries1, tri_num1
    else:
        autocorr = False
        trial_boundaries2, tri_num2 = _get_trial_boundaries(spk, cell_idx2)

    idx2 = 0
    for idx1, tri in enumerate(tri_num1):

        if not autocorr:
            # check if the other unit has spikes in this trial
            has_tri = tri in tri_num2[idx2:]
            if has_tri:
                idx2 = np.where(tri_num2[idx2:] == tri)[0][0] + idx2
            else:
                continue

        times1 = spk.time[cell_idx1][
            trial_boundaries1[idx1]:trial_boundaries1[idx1 + 1]]

        if autocorr:
            times2 = times1
        else:
            times2 = spk.time[cell_idx2][
                trial_boundaries2[idx2]:trial_boundaries2[idx2 + 1]]

        time_diffs = times2[:, None] - times1[None, :]

        if autocorr:
            n_diffs = time_diffs.shape[0]
            ind = np.diag_indices(n_diffs)
            time_diffs[ind] = np.nan

        time_diffs = time_diffs.ravel()
        this_hist, _ = np.histogram(time_diffs, bins=bins)
        xcorr[tri, :] = this_hist

    return xcorr, bins


# TODO: max_lag cannot be longer than epoch time window and preferrably
#       it is < 1 / 2 epoch length ...
# TODO: when gauss_fwhm is set, calculate trim before and adjust max_lag
#       so that it is left in after convolution
def xcorr_hist_trials(spk, picks=None, picks2=None, sfreq=500., max_lag=0.2,
                      bins=None, gauss_fwhm=None):
    '''Compute cross-correlation histogram for each trial.

    Parameters
    ----------
    spk : SpikeEpochs
        SpikeEpochs to compute cross-correlation for.
    picks1 : ...
        List of cell indices or names to perform cross- and auto- correlations
        for. All combinations of cells will be used.
    picks2 : ...
        ...
    sfreq : float
        Sampling frequency of the bins. The bin width will be ``1 / sfreq``
        seconds. Used only when ``bins is None``. Defaults to ``500.``.
    max_lag : float
        Maximum lag in seconds. Used only when ``bins is None``. Defaults
        to ``0.2``.
    bins : numpy array | None
        Array representing edges of the histogram bins. If ``None`` (default)
        the bins are constructed based on ``sfreq`` and ``max_lag``.
    gauss_fwhm : float | None
        Full-width at half maximum of the gaussian kernel to convolve the
        cross-correlation histograms with. Defaults to ``None`` which ommits
        convolution.

    Returns
    -------
    xcorr : xarray.DataArray
        ...
    '''
    from .spikes import SpikeEpochs
    from .utils import _deal_with_picks, _turn_spike_rate_to_xarray
    assert isinstance(spk, SpikeEpochs)

    bins = _construct_bins(sfreq, max_lag, bins=bins)

    picks = _deal_with_picks(spk, picks)
    if picks2 is None:
        # picks and picks2 has to be all combinations of picks
        pick_pairs = itertools.product(picks, picks)
    else:
        picks2 = _deal_with_picks(spk, picks2)
        pick_pairs = itertools.product(picks, picks2)

    cell = list()  # pair name, but set as cell attr in xarray
    cell1_idx, cell2_idx = list(), list()
    cell1_name = list()
    cell2_name = list()
    auto = list()  # is autocorrelation

    # LOOP
    xcorrs = list()
    for idx1, idx2  in pick_pairs:
        # compute xcorr histogram for cell pair
        xcorr, _ = _xcorr_hist_trials(spk, idx1, idx2, sfreq=sfreq,
                                      max_lag=max_lag, bins=bins)

        name1 = spk.cell_names[idx1]
        name2 = spk.cell_names[idx2]
        is_auto = idx1 == idx2
        name = f'{name1} x {name2}'

        cell.append(name)
        cell1_name.append(name1)
        cell2_name.append(name2)
        cell1_idx.append(idx1)
        cell2_idx.append(idx2)

        xcorrs.append(xcorr)

    # stack cell pairs as first dimension
    xcorrs = np.stack(xcorrs, axis=0)

    # calc bin_centers
    bin_widths = np.diff(bins)
    bin_centers = bins[:-1] + bin_widths

    # smooth with gaussian if needed
    # TODO: DRY with _spike_density
    if gauss_fwhm is not None:
        from scipy.signal import correlate
        from .spike_rate import _gauss_sd_from_FWHM

        gauss_sd = _gauss_sd_from_FWHM(gauss_fwhm)
        winlen = gauss_sd * 6
        gauss_sd = gauss_sd * sfreq
        win_smp, trim = _symmetric_window_samples(winlen, sfreq)
        kernel = _gauss_kernel_samples(win_smp, gauss_sd) * sfreq

        bin_centers = bin_centers[trim:-trim]
        xcorrs = correlate(xcorrs, kernel[None, None, :], mode='valid')

    # construct xarr
    xcorrs = _turn_spike_rate_to_xarray(
        bin_centers, xcorrs, spk, cell_names=cell, x_dim_name='lag')
    xcorrs.name = 'count'
    xcorrs.attrs['coord_units'] = {'lag': 's'}

    # add cell1_idx etc.
    xcorrs = xcorrs.assign_coords(
        {'cell1_name': ('cell', cell1_name),
         'cell2_name': ('cell', cell2_name),
         'cell1_idx': ('cell', cell1_idx),
         'cell2_idx': ('cell', cell2_idx)}
    )

    return xcorrs


def _spike_xcorr_density(spk, cell_idx, picks=None, sfreq=500, winlen=0.1,
                         kernel_winlen=0.025):
    from .spike_rate import _spike_density

    # create kernel
    gauss_sd = kernel_winlen / 6 * sfreq
    win_smp, trim = _symmetric_window_samples(kernel_winlen, sfreq)
    kernel = _gauss_kernel_samples(win_smp, gauss_sd) * sfreq

    # calculate spike density
    picks = _deal_with_picks(spk, picks)
    tms, cnt = _spike_density(spk, picks=picks, sfreq=sfreq, kernel=kernel)

    # cut out spike-centered windows
    windows = spike_centered_windows(
        spk, cell_idx, cnt, tms, sfreq, winlen=winlen)

    # correct autocorrelation if present:
    if cell_idx in picks:
        idx = np.where(np.asarray(picks) == cell_idx)[0][0]
        trim = int((windows.shape[-1] - len(kernel)) / 2)
        windows[idx, :, trim:-trim] -= kernel

    # turn to xarray
    t_per_smp = 1 / sfreq
    win_diff = [-winlen / 2, winlen / 2]
    time = np.arange(win_diff[0], win_diff[1] + 0.01 * t_per_smp,
                     step=t_per_smp)
    cell_names = [spk.cell_names[idx] for idx in picks]
    windows = windows.transpose('channel', 'spike', 'time')
    xcorr = _turn_spike_rate_to_xarray(
        time, windows, spk, tri=windows.trial.values,
        cell_names=cell_names
    )

    return xcorr


# - [ ] silence warnings
# - [ ] add more arguments that could be passed to elephant
# - [ ]
def _spike_xcorr_elephant(spk, cell_idx1, cell_idx2, sfreq=500, winlen=0.1,
                          kernel_winlen=0.025, shift_predictor=False):
    from scipy.signal import correlate
    import quantities as pq
    from elephant.conversion import BinnedSpikeTrain
    from elephant.spike_train_correlation import cross_correlation_histogram

    # create kernel
    gauss_sd = kernel_winlen / 6 * sfreq
    win_smp, trim = _symmetric_window_samples(kernel_winlen, sfreq)
    kernel = _gauss_kernel_samples(win_smp, gauss_sd) * sfreq

    # bin spikes
    binsize = 1 / sfreq
    spk1 = spk.to_neo(cell_idx1)
    spk2 = spk.to_neo(cell_idx2)
    bst1 = BinnedSpikeTrain(spk1, bin_size=binsize * pq.s)
    bst2 = BinnedSpikeTrain(spk2, bin_size=binsize * pq.s)

    # window length
    win_len = int(winlen / 2 / binsize)

    cch_list = list()
    n_tri = bst1.shape[0]
    if shift_predictor:
        n_tri -= 1

    # iterate through trials
    for tri in range(n_tri):
        if not shift_predictor:
            tri1, tri2 = tri, tri
        else:
            tri1, tri2 = tri, tri + 1

        cch, lags = cross_correlation_histogram(
            bst1[tri1], bst2[tri2], window=[-win_len, win_len], kernel=kernel)
        cch_list.append(np.array(cch)[:, 0])

    # add last trial if shift predictor
    if shift_predictor:
        cch, lags = cross_correlation_histogram(
            bst1[-1], bst2[-2], window=[-50, 50], kernel=kernel)
        cch_list.append(np.array(cch)[:, 0])

    cch_list = np.stack(cch_list, axis=0)
    lags = lags * binsize

    cell_name = '{}-{}'.format(spk.cell_names[cell_idx1],
                               spk.cell_names[cell_idx2])
    cch = _turn_spike_rate_to_xarray(
        lags, cch_list[None, :], spk, cell_names=[cell_name],
        copy_cellinfo=False)
    return cch


def xcorr_signal(x, y, maxlag=None):
    '''Compute un-normalized cross-correlation for required max lag.

    This is equivalent to convolution but without reversing the kernel.

    Parameters
    ----------
    x : numpy.ndarray
        First 1d array to crosscorrelate.
    y : numpy.ndarray
        Second 1d array to crosscorrelate.
    maxlab : int
        Maximum lag in samples.

    Returns
    -------
    lags : numpy.ndarray
        1d array of lags (in samples).
    corr : numpy.ndarray
        1d array of crosscorrelations.
    '''
    from scipy.signal import correlate, correlation_lags

    corr = correlate(x, y)
    lags = correlation_lags(len(x), len(y))

    if maxlag is not None:
        from borsar.utils import find_range

        rng = find_range(lags, [-maxlag, maxlag])
        corr, lags = corr[rng], lags[rng]

    return lags, corr


def shuffled_spike_xcorr(spk, cell_idx1, cell_idx2, sfreq=500,
                         gauss_winlen=0.025, max_lag=0.1, pbar=True,
                         n_shuffles=1_000):
    '''FIXME'''
    from scipy.signal import correlate
    from .viz import check_modify_progressbar

    pbar = check_modify_progressbar(pbar, total=n_shuffles)
    max_lag_smp = int(np.round(sfreq * max_lag))

    time, bin_spk = spk.to_raw(picks=[cell_idx1, cell_idx2], sfreq=sfreq)
    n_trials, n_cells, n_times = bin_spk.shape

    # add maxlag separation between trials
    add_sep = np.zeros(shape=(n_trials, n_cells, max_lag_smp))
    bin_spk = np.concatenate([bin_spk, add_sep], axis=2)
    n_times = bin_spk.shape[-1]

    win, _ = _symmetric_window_samples(gauss_winlen, sfreq=sfreq)
    sd_smp = int(np.round((sfreq * gauss_winlen) / 6))
    gauss = _gauss_kernel_samples(win, sd_smp)

    # unroll to cells x (trials * times)
    concat = bin_spk.transpose(1, 0, 2).reshape(n_cells, n_trials * n_times)

    # turn spikes into continuous signal
    cnt = correlate(concat, gauss[None, :], mode='same')
    lags, corr = xcorr_signal(cnt[0], cnt[1], maxlag=max_lag_smp)
    lags = lags / sfreq

    if n_shuffles > 0:
        n_lags = len(lags)
        corr_shuffled = np.zeros(shape=(n_shuffles, n_lags))

        cnt1 = cnt[0]
        # shuffle trials n_times
        for shuffle_idx in range(n_shuffles):
            tri = np.arange(n_trials)
            np.random.shuffle(tri)

            # unroll to cells x (trials * times)
            bin_spk_shuffled = bin_spk[tri, 1]
            concat_shuffled = bin_spk_shuffled.reshape(n_trials * n_times)

            # turn spikes into continuous signal
            cnt_shuffled = correlate(concat_shuffled, gauss, mode='same')
            _, this_corr = xcorr_signal(cnt1, cnt_shuffled, maxlag=max_lag_smp)
            corr_shuffled[shuffle_idx] = this_corr
            pbar.update(1)

        avg_shfl = corr_shuffled.mean(axis=0)
        std_shfl = corr_shuffled.std(axis=0)
    else:
        avg_shfl = 0
        std_shfl = 1

    return lags, (corr - avg_shfl) / std_shfl


def find_coincidence_clusters(similarity, threshold=0.3):
    import borsar

    adj = similarity >= threshold
    suspicious = adj.any(axis=0) | adj.any(axis=1)
    suspicious_idx = np.where(suspicious)[0]
    adj_susp = adj[suspicious_idx[:, None], suspicious_idx[None, :]]

    fake_signal = np.ones(len(suspicious_idx))
    clusters, counts = borsar.cluster.find_clusters(
        fake_signal, 0.5, adjacency=adj_susp, backend='mne')
    return suspicious_idx, clusters, counts


# TODO: clean up zasady comments
# TODO: clean up a bit, many indexing levels make it a bit confusing to read
#       and modify
# TODO: some time, maybe add more control (kwargs)
# TODO: allow to drop one from pair when different channels, different
#       alignments, high waveform similarity (?) - in case of ground ref this
#       may happen...
def drop_duplicated_units(spk, similarity, return_clusters=False,
                          verbose=False, different_alignment_threshold=0.4,
                          different_channel_threshold=0.4):
    # %% zasady
    # podobieństwo 1.0, ta sama elektroda -> duplikat, usuwamy dowolny
    #
    # jeżeli ten sam kanał b duże podobienstwo (np. > 0.5), różny alignment
    # - wybierz neuron, który ma więcej spike'ów? (czyli de facto - wybierz
    # ten, którego mniej spike'ów zawiera się w drugim)

    # duży klaster, wiele elektrod, wysokie podobieństwa (ok > 0.4)
    # oraz wiele par ma wysoką korelację waveformów ->
    # zrobić osort ze średnia referencją
    #
    # inny kanał, bardzo podobny średni waveform (np. cross-corr > 0.9) ->
    # wybierz ten co ma więcej spike'ow
    #
    # gdy tylko kilka neuronów w klastrze:
    # różne kanały, >= 30% ko-spike'ów, wysoki peak kroskorelacji średnich
    # wavefromow -> wybierz ten, co ma więcej spike'ów
    import scipy

    # init drop vector
    drop = np.zeros(len(spk.timestamps), dtype='bool')

    # cluster by similarity over 0.3 threshold
    threshold = min(different_alignment_threshold, different_channel_threshold)
    suspicious_idx, clusters, counts = find_coincidence_clusters(
        similarity, threshold=threshold)

    # go through pairs within each cluster and apply selection rules
    for cluster_idx in range(len(clusters)):
        idxs = suspicious_idx[clusters[cluster_idx]]
        simil_part = similarity[idxs[:, None], idxs[None, :]]

        # 1. remove duplicates (any similarity == 1.)
        msg = 'Removed {}-{} pair - identical units.'
        if (simil_part == 1).any():
            s1, s2 = np.where(simil_part == 1.)
            for ix in range(len(s1)):
                identical = min([s1[ix], s2[ix]])
                drop[idxs[identical]] = True

                if verbose:
                    name1 = spk.cell_names[idxs[s1[ix]]]
                    name2 = spk.cell_names[idxs[s2[ix]]]
                    print(msg.format(name1, name2))

        # 2. pairs with similarity >= 0.4, same channels, different alignment
        info = spk.cellinfo.loc[idxs, :]
        if (simil_part >= different_alignment_threshold).any():
            msg = ('Removed {}-{} pair - same channel, different alignment, '
                   'spike coincidence: {:.3f}.')
            s1, s2 = np.where(simil_part > different_alignment_threshold)
            for ix in range(len(s1)):
                ix1, ix2 = s1[ix], s2[ix]
                same_chan = info.channel.iloc[ix1] == info.channel.iloc[ix2]
                same_align = (info.alignment.iloc[ix1]
                              == info.alignment.iloc[ix2])
                if same_chan and not same_align:
                    s1_in_s2 = simil_part[ix1, ix2]
                    s2_in_s1 = simil_part[ix2, ix1]
                    rel_idx = np.argmax([s1_in_s2, s2_in_s1])
                    drop_idx = [ix1, ix2][rel_idx]
                    drop[idxs[drop_idx]] = True

                    if verbose:
                        name1 = spk.cell_names[idxs[ix1]]
                        name2 = spk.cell_names[idxs[ix2]]
                        this_simil = max([s1_in_s2, s2_in_s1])
                        print(msg.format(name1, name2, this_simil))

        # 2. pairs with similarity > 0.3, different channels,
        #    very similar waveforms

        msg = ('Removed {}-{} pair - different channel, very similar waveform, '
                'spike coincidence: {:.3f}, max waveform xcorr: {:.3f}.')
        s1, s2 = np.where(simil_part >= different_channel_threshold)
        avg_wave = [spk.waveform[idx].mean(axis=0) for idx in idxs]
        for ix in range(len(s1)):
            ix1, ix2 = s1[ix], s2[ix]
            same_chan = info.channel.iloc[ix1] == info.channel.iloc[ix2]
            same_align = info.alignment.iloc[ix1] == info.alignment.iloc[ix2]
            if not same_chan:
                s1_in_s2 = simil_part[ix1, ix2]
                s2_in_s1 = simil_part[ix2, ix1]
                rel_idx = np.argmax([s1_in_s2, s2_in_s1])
                drop_idx = [ix1, ix2][rel_idx]

                # but only if waveform max cross-corr is > 0.9
                wave1 = avg_wave[ix1]
                wave2 = avg_wave[ix2]
                corr = scipy.signal.correlate(wave1, wave2)
                auto1 = (wave1 * wave1).sum()
                auto2 = (wave2 * wave2).sum()
                norm = np.sqrt(auto1 * auto2)
                corr = corr / norm
                corr_maxabs = np.abs(corr).max()

                if corr_maxabs > 0.99:
                    drop[idxs[drop_idx]] = True

                    if verbose:
                        name1 = spk.cell_names[idxs[ix1]]
                        name2 = spk.cell_names[idxs[ix2]]
                        this_simil = max([s1_in_s2, s2_in_s1])
                        print(msg.format(name1, name2, this_simil, corr_maxabs))

    if not return_clusters:
        return drop
    else:
        return drop, clusters, counts, suspicious_idx


# TODO: move to viz
def plot_high_similarity_cluster(spk, similarity, clusters, suspicious_idx,
                                 cluster_idx=0, drop=None, figsize=(14, 9)):
    '''Plot similarity matrix with waveforms in top column and leftmost rows.

    Parameters
    ----------
    spk : Spikes | SpikeEpochs
        Object containing spikes.
    similarity : numpy.array
        Numpy array with coincidence similarity.
    clusters : list of numpy.array
        List of boolean arrays - where each array identifies cluster members
        with ``True`` values.
    suspicious_idx : numpy.array
        Array of indices mapping from elements in clusters to indices of
        units in ``spk`` (FIXME - better description).
    cluster_idx : int
        Which cluster to plot. Defaults to ``0`` (the first cluster).
    drop : numpy.array | None
        Boolean array specifying which units will be dropped. Optional,
        is used to color unit titles, defaults to ``None``, when the titles are
        not colored in red.
    figsize : tuple
        Two-element tuple specifying the size of the figure (in inches, as
        matploltib likes it).

    Returns
    -------
    fig : matplotlib.Figure
        Figure object.
    '''
    import matplotlib.pyplot as plt

    idxs = suspicious_idx[clusters[cluster_idx]]

    n_cells = len(idxs)
    simil_part = similarity[idxs[:, None], idxs[None, :]]

    fig, ax = plt.subplots(nrows=n_cells + 1, ncols=n_cells + 1,
                           figsize=figsize)

    title_fontsize = (12 if n_cells < 9 else 10 if n_cells < 13
                      else 8 if n_cells < 17 else 5)

    for idx, cell_idx in enumerate(idxs):
        spk.plot_waveform(cell_idx, ax=ax[0, idx + 1])
        spk.plot_waveform(cell_idx, ax=ax[idx + 1, 0])

        info = spk.cellinfo.loc[cell_idx, :]
        n_spikes = len(spk.timestamps[cell_idx])
        alg_txt = (f'\n{info.alignment}'
                   if not info.alignment == 'unknown'
                   else '')
        title = (f'{info.channel}\ncluster {info.cluster}' + alg_txt +
                 f'\n{n_spikes} spikes')
        color = (('red' if drop[cell_idx] else 'black') if drop is not None
                 else 'black')

        ax[0, idx + 1].set_title(title, color=color, fontsize=title_fontsize)

    for this_ax in ax.ravel():
        this_ax.set_ylabel('')
        this_ax.set_xlabel('')
        this_ax.set_xticks([])
        this_ax.set_yticks([])

    ax[0, 0].axis(False)
    max_val = simil_part.max()
    for row_idx in range(n_cells):
        for col_idx in range(n_cells):
            if row_idx == col_idx:
                ax[row_idx + 1, col_idx + 1].axis(False)
                continue

            value = simil_part[row_idx, col_idx]
            val_perc = value / max_val
            txt_col = 'black' if val_perc > 0.5 else 'white'
            ax[row_idx + 1, col_idx + 1].text(0.5, 0.5, f'{value:0.3f}',
                                            horizontalalignment='center',
                                            verticalalignment='center',
                                            color=txt_col)

            color = plt.cm.viridis(val_perc)
            ax[row_idx + 1, col_idx + 1].set_facecolor(color)

    return fig



# -*- coding: utf-8 -*-
"""
Created on Wed Oct 11 14:55:37 2023

@author: mmagnuski
"""
import os.path as op

import numpy as np
import pandas as pd
from scipy.io import loadmat

import pylabianca as pln


# we read matlab file with events
data_dir = r'C:\Users\Asus\Dropbox\Sarenka\PROJECTS\pylabianca\ft_data'
event_fname = 'p029_sort_final_01_events.mat'
events = pln.io.read_events_plexon_nex(op.join(data_dir, event_fname))

# and then format the events into mne-python events array
event_time = events[:, 0]
event_id = events[:, -1]


# %%
# target changes first: 20001, 20003
# distractor changes first: 20002, 20004
# # so attention in RF: 20001 and 20002
# and attention out of RF is 20003, 20004

events = {'begin trial': 10044, 'end trial': 10045, 'stim on': 10030,
          'target change': 12001, 'distractor change': 12000,
          'correct response': 10041}
attCnds = np.arange(20001, 20004 + 1)

begmark = np.where(event_id == events['begin trial'])[0]
endmark = np.where(event_id == events['end trial'])[0]

assert len(begmark) == len(endmark)
assert (endmark > begmark).all()

columns=['has_stimon', 'has_targetchange', 'has_distractorchange',
         'condition', 'correct', 'change_first', 'attention']
df = pd.DataFrame(columns=columns)

for idx in range(len(begmark)):
    trigs = event_id[begmark[idx]:endmark[idx] + 1]
    trig_ts = event_time[begmark[idx]:endmark[idx] + 1]
    t_start = trig_ts[0] / 40_000

    which_stimon = np.where(trigs == events['stim on'])[0]
    n_stm = len(which_stimon)
    has_stimon = n_stm > 0
    df.loc[idx, 'has_stimon'] = has_stimon

    tchange_idx = np.where(trigs == events['target change'])[0]
    dchange_idx = np.where(trigs == events['distractor change'])[0]

    df.loc[idx, 'has_targetchange'] = len(tchange_idx) > 0
    df.loc[idx, 'has_distractorchange'] = len(dchange_idx) > 0

    # condition
    cond_idx = np.where(np.isin(trigs, attCnds))[0][0]
    cond_val = trigs[cond_idx]
    change_first = 'target' if (cond_val % 2) == 1 else 'distractor'
    att_cnd = 'in' if cond_val < 20003 else 'out'
    is_correct = events['correct response'] in trigs
    df.loc[idx, 'condition'] = cond_val
    df.loc[idx, 'change_first'] = change_first
    df.loc[idx, 'attention'] = att_cnd
    df.loc[idx, 'correct'] = is_correct

# %%
df = df.infer_objects()
df.to_csv(op.join(data_dir, 'monkey_stim.csv'), index=False)


import numpy as np
import pandas as pd
import pylabianca as pln


def create_fake_spikes():
    times = [[-0.3, -0.1, 0.025, 0.11, 0.22, 0.25, 0.4,
              -0.08, 0.12, 0.14, 0.19, 0.23, 0.32],
             [-0.22, -0.13, -0.03, 0.08, 0.16, 0.33, -0.2,
              -0.08, 0.035, 0.148, 0.32]]
    trials = [[0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1],
              [0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1]]
    spk = pln.SpikeEpochs(times, trials)
    return spk


def create_random_spikes(n_cells=4, **args):
    n_trials = 25
    tmin, tmax = -0.5, 1.5
    tlen = tmax - tmin
    n_spikes_per_tri = [10, 21]

    times = list()
    trials = list()
    for cell_idx in range(n_cells):
        this_tri = list()
        this_tim = list()
        for tri_idx in range(n_trials):
            n_spk = np.random.randint(*n_spikes_per_tri)
            tms = np.random.rand(n_spk) * tlen + tmin
            this_tri.append(np.ones(n_spk) * tri_idx)
            this_tim.append(tms)
        this_tri = np.concatenate(this_tri)
        this_tim = np.concatenate(this_tim)
        times.append(this_tim)
        trials.append(this_tri)
    return pln.SpikeEpochs(times, trials, **args)


def test_crop():
    spk = create_fake_spikes()
    spk.crop(tmin=0.1)

    assert (spk.time[0] == np.array(
        [0.11, 0.22, 0.25, 0.4, 0.12, 0.14, 0.19, 0.23, 0.32])).all()
    assert (spk.time[1] == np.array([0.16, 0.33, 0.148, 0.32])).all()
    assert (spk.trial[0] == np.array([0, 0, 0, 0, 1, 1, 1, 1, 1])).all()
    assert (spk.trial[1] == np.array([0, 0, 1, 1])).all()


def test_repr():
    spk = create_fake_spikes()
    should_be = '<SpikeEpochs, 2 epochs, 2 cells, 12.0 spikes/cell on average>'
    assert str(spk) == should_be


def test_pick_cells():
    spk = create_fake_spikes()
    assert len(spk.time) == 2
    spk.pick_cells('cell001')
    assert len(spk.time) == 1
    assert spk.time[0][0] == -0.22


def test_pick_cells_cellinfo_query():
    from copy import deepcopy
    cellinfo = pd.DataFrame({'cell_idx': [10, 15, 20, 25],
                             'letter': list('abcd')})
    spk = create_random_spikes(cellinfo=cellinfo)

    spk2 = deepcopy(spk)
    spk2.pick_cells(query='cell_idx > 18')
    assert len(spk2.time) == 2
    assert spk2.cellinfo.shape[0] == 2
    assert (spk2.cellinfo.letter.values == np.array(['c', 'd'])).all()

    spk3 = deepcopy(spk)
    spk3.pick_cells(query="letter in ['a', 'c']")
    assert len(spk3.time) == 2
    assert len(spk3.time[1]) == len(spk.time[2])
    assert (spk3.cellinfo.cell_idx.values == [10, 20]).all()


def test_to_raw():
    times = [[-0.3, -0.28, -0.26, 0.15, 0.18, 0.2],
             [-0.045, 0.023, -0.1, 0.13]]
    trials = [[0, 0, 0, 0, 0, 0], [0, 0, 1, 1]]
    spk = pln.SpikeEpochs(times, trials)

    spk_tm, spk_raw = pln.spikes._spikes_to_raw(spk, sfreq=10)
    print(spk_raw)
    good_tms = np.arange(-0.3, 0.21, step=0.1)
    good_raw = np.array(
        [[[3, 0, 0, 0, 1, 2], [0, 0, 0, 2, 0, 0]],
         [[0, 0, 0, 0, 0, 0], [0, 0, 1, 0, 1, 0]]]
    )
    assert (spk_tm == good_tms).all()
    assert (spk_raw == good_raw).all()



import os
import os.path as op
from pathlib import Path
from warnings import warn

import numpy as np
import pandas as pd

from .spikes import SpikeEpochs, Spikes


# TODO - trialinfo columns ...
def read_filedtrip(fname, data_name='spike', kind='raw', waveform=True):
    '''Read fieldtrip SpikeTrials format.

    Parameters
    ----------
    fname : str | pathlib.Path
        Path to the file to read.
    data_name : str
        The name of the saved variable - this can be arbitrary so it is
        necessary to specify. ``'spike'`` by default.
    kind : str
        Data format to read. Can be:
        * ``'raw'`` for FieldTrip raw spikes format
        * ``'trials'`` or ``'epochs'`` for FieldTrip SpikeTrials format
    waveform : bool
        Whether to read waveforms. Defaults to ``True``.

    Returns
    -------
    spk : SpikeEpochs
        SpikeEpochs object.
    '''
    from scipy.io import loadmat

    accept_kind = ['epochs', 'trials', 'raw']
    if kind not in accept_kind:
        msg = (f'`kind` has to be one of {accept_kind}, got {kind}.')
        raise ValueError(msg)

    data = loadmat(fname, squeeze_me=True, variable_names=data_name)[data_name]

    cell_names = data['label'].item()

    if waveform:
        waveform, waveform_time = _get_waveforms(data)
    else:
        waveform, waveform_time = None, None

    if 'trialinfo' in data.dtype.names:
        trialinfo = data['trialinfo'].item()
        trialinfo = pd.DataFrame(trialinfo)
    else:
        trialinfo = None

    if 'cellinfo' in data.dtype.names:
        data_dct = dict()
        fields = data['cellinfo'].item().dtype.names
        for fld in fields:
            data_dct[fld] = data['cellinfo'].item()[fld].item()
        cellinfo = pd.DataFrame(data_dct)
    else:
        cellinfo = None

    if kind in ['trials', 'epochs']:
        spk = _read_ft_spikes_tri(data, cell_names, trialinfo, cellinfo,
                                  waveform, waveform_time)
    elif kind == 'raw':
        spk = _read_ft_spikes_raw(data, cell_names, cellinfo,
                                  waveform, waveform_time)

    spk.filename = fname
    return spk


# TODO
# - [ ] read waveform too...
def _read_ft_spikes_tri(data, cell_names, trialinfo, cellinfo, waveform,
                        waveform_time):
    '''Read fieldtrip SpikeTrials format.

    Returns
    -------
    spk : SpikeEpochs
        SpikeEpochs object.
    '''

    time = data['time'].item()
    trial = data['trial'].item() - 1
    trialtime = data['trialtime'].item()

    msg = 'All trials have to be of the same length'
    assert (trialtime == trialtime[[0]]).all(), msg

    n_trials = trialtime.shape[0]
    time_limits = trialtime[0]

    # create SpikeEpochs
    spk = SpikeEpochs(time, trial, time_limits, n_trials=n_trials,
                      metadata=trialinfo, cell_names=cell_names,
                      cellinfo=cellinfo, waveform=waveform,
                      waveform_time=waveform_time)

    return spk


def _read_ft_spikes_raw(data, cell_names, cellinfo, waveform,
                        waveform_time):
    '''Read raw spikes fieldtrip format.

    Returns
    -------
    spk : Spikes
        Spikes object.

    '''
    timestamps = data['timestamp'].item()
    timestamps = _check_timestamps(timestamps)
    sfreq = data['hdr'].item()['FileHeader'].item()['Frequency'].item()

    # create Spikes
    spk = Spikes(timestamps, sfreq, cell_names=cell_names,
                 cellinfo=cellinfo, waveform=waveform,
                 waveform_time=waveform_time)
    return spk


def _check_timestamps(timestamps):
    from numbers import Integral

    n_units = len(timestamps)
    for ix in range(n_units):
        if isinstance(timestamps[ix], Integral):
            timestamps[ix] = np.array([timestamps[ix]])

    return timestamps


def _get_waveforms(data):
    '''Get waveforms from a FieldTrip file.'''
    sfreq = data['hdr'].item()['FileHeader'].item()['Frequency'].item()

    if 'waveform' in data.dtype.names:
        waveforms = data['waveform'].item()
        new_waveforms = list()
        n_units = len(waveforms)
        for ix in range(n_units):
            this_waveform = waveforms[ix]

            if (isinstance(this_waveform, np.ndarray)
                and len(this_waveform) > 0):

                # we currently take only the first lead
                if this_waveform.ndim > 2:
                    this_waveform = this_waveform[0]
                elif this_waveform.ndim == 1:
                    # only one waveform, squeezed
                    this_waveform = this_waveform[:, None]
                new_waveforms.append(this_waveform.T)
            else:
                new_waveforms.append(None)

        waveform_length = [x.shape[1] for x in new_waveforms if x is not None]
        same_lengths = all(waveform_length[0] == x
                           for x in waveform_length[1:])

        if same_lengths:
            # waveform time in ms
            waveform_time = np.arange(waveform_length[0]) / (sfreq / 1000)
        else:
            warn('Not all waveforms have the same number of samples, '
                 'waveforms are therefore ignored. Got the following waveform'
                 f'lengths: {waveform_length}.')
            waveforms, waveform_time = None, None
    else:
        waveforms, waveform_time = None, None

    return new_waveforms, waveform_time


# TODO: add progressbar?
def read_combinato(path, label=None, alignment='both'):
    '''Read spikes from combinato sorting output.

    Note that groups are read as single units, the class information is stored
    inside the Spikes object in `.cellinfo` attribute.

    Parameters
    ----------
    path : str
        Path to directory with combinato channel subdirectories with sorting
        results.
    label : str | None
        Read specific sorting labels (if multiple). Defaults to ``None``, which
        reads the first available label.
    alignment : str
        The alignment to read. Can be:
        * ``'pos'`` - read only positive alignment sorting results
        * ``'neg'`` - read only negative alignment sorting results
        * ``'both'`` - read both alignments
    '''
    import h5py

    if alignment == 'both':
        alignment = ['neg', 'pos']
    elif isinstance(alignment, str):
        alignment = [alignment]

    # currently we read only SU by default
    types_oi = [2]  # 2: SU; 1: MU; 0: artifact

    channel_dirs = os.listdir(path)
    group = list()
    align = list()
    channel = list()
    spike_data = {'timestamp': list(), 'waveform': list(), 'class': list(),
                  'distance': list()}

    for subdir in channel_dirs:
        has_content = False
        subdir_path = op.join(path, subdir)
        hdf5_file = op.join(subdir_path, 'data_' + subdir + '.h5')

        if op.isdir(subdir_path) and op.exists(hdf5_file):
            sort_dirs = os.listdir(subdir_path)

            for pol in alignment:
                subdir_prefix = f'sort_{pol}_'
                pol_subdirs = [f for f in sort_dirs if subdir_prefix in f]

                if len(pol_subdirs) == 0:
                    continue

                if label is None:
                    labels = np.unique([dr.split('_')[2]
                                       for dr in pol_subdirs])
                    if labels.shape[0] > 1:
                        msg = ('Multiple sorting labels present, please provi'
                               'de a sorting label to read results from. Found'
                               f' the following labels in {subdir} directory: '
                                ', '.join(labels))
                        raise RuntimeError(msg)
                    else:
                        label = str(labels[0])

                pol_subdir = f'sort_{pol}_{label}'
                full_sorting_path = op.join(subdir_path, pol_subdir,
                                            'sort_cat.h5')
                if not op.exists(full_sorting_path):
                    continue

                sorting_file = h5py.File(full_sorting_path, 'r')

                types = np.asarray(sorting_file['types'])
                # find SUs (or SUs and MUs)
                is_SU = np.in1d(types[:, -1], types_oi)

                if not is_SU.any():
                    continue

                if not has_content:
                    spikes_file = h5py.File(hdf5_file, 'r')

                has_content = True
                groups_oi = types[is_SU, 0]
                groups = np.asarray(sorting_file['groups'])
                groups_sel = np.in1d(groups[:, 1], groups_oi)
                groups = groups[groups_sel, :]

                spike_classes = np.asarray(sorting_file['classes'])
                spike_indices = np.asarray(sorting_file['index'])
                spike_distance = np.asarray(sorting_file['distance'])
                sorting_file.close()

                times = np.asarray(spikes_file[pol]['times'])
                waveforms = np.asarray(spikes_file[pol]['spikes'])

                for grp in groups_oi:
                    msk = groups[:, 1] == grp
                    this_classes = groups[msk, 0]
                    class_msk = np.in1d(spike_classes, this_classes)

                    idx = spike_indices[class_msk]
                    # some groups can be empty with label that was not updated
                    if len(idx) > 0:
                        group.append(grp)
                        align.append(pol)
                        channel.append(subdir)

                        # Combinato times are in ms, but we turn this
                        # to 1 microsecond timestamps used by Neuralynx
                        spike_data['timestamp'].append(times[idx] * 1000)
                        spike_data['waveform'].append(waveforms[idx, :])

                        spike_data['class'].append(spike_classes[class_msk])
                        spike_data['distance'].append(
                            spike_distance[class_msk])

        if has_content:
            spikes_file.close()

    # organize into one Spikes object
    cellinfo = pd.DataFrame(data={'channel': channel, 'alignment': align,
                                  'group': group})
    spikes = Spikes(spike_data['timestamp'], sfreq=1e6, cellinfo=cellinfo,
                    waveform=spike_data['waveform'])
    return spikes


def read_osort(path, waveform=True, channels='all', format='mm',
               progress=True):
    '''Read osort sorting results.

    The mm format can be obtained using updateSORTINGresults_mm matlab
    function from _psy_screening (located in
    ``psy_screenning-/helpers/sorting_utils``).

    Parameters
    ----------
    path : str
        Path to the directory with the ``.mat`` files obtained with
        ``updateSORTINGresults_mm``. It can also be one file with all the units
        in it.
    waveform : bool
        Whether to also read the waveforms. Waveforms typically take the most
        memory so it may be worth turning it off for fast reading. Defaults
        to ``True``.
    channels : str | list
        Channels to read. Defaults to ``'all'``, which reads all channels.
    format : str
        Osort file format. Can be:
        * ``'standard'`` - the default osort format
        * ``'mm'`` - format with cleaned up variable names and with alignment
            information
    progress : bool
        Whether to show progress bar. Defaults to ``True``.

    Returns
    -------
    spk : pylabianca.spikes.Spikes
        Spikes object.
    '''
    from scipy.io import loadmat
    if progress:
        from tqdm import tqdm

    assert op.exists(path), 'Path/file does not exist.'
    assert format in ['standard', 'mm'], 'Unknown format.'
    one_file = not op.isdir(path)

    if not one_file:
        files = [f for f in os.listdir(path) if f.endswith('.mat')]
        files.sort()

        # select files based on channels and format
        if not channels == 'all':
            if isinstance(channels, str):
                channels = [channels]
            channels_check = [ch + '_' for ch in channels]
            check_channel = lambda f: any([ch in f for ch in channels_check])
            files = [f for f in files if check_channel(f)]
        if format == 'mm':
            files = [f for f in files if 'mm_format' in f]
    else:
        full_path = Path(path)
        path = full_path.parent
        files = [full_path.name]

    cluster_id, alignment, threshold, channel = [list() for _ in range(4)]
    timestamp = list()
    waveforms = list() if waveform else None
    ignore_cluster = [0, 1, 99999999]

    if format == 'mm':
        # TEMP FIX: older exporting function had a spelling error
        correct_field = 'alignment' if one_file else ['aligment', 'alignment']
        var_names = None
        read_vars = ['cluster_id', 'threshold', 'channel', 'timestamp']
        if waveform:
            read_vars.append('waveform')

        translate = {var: var for var in read_vars}
        if one_file:
            var_names = read_vars + [correct_field]
    else:
        var_names = ['assignedNegative', 'newTimestampsNegative']
        # var_names = ['assignedNegative', '', '', 'newTimestampsNegative']
        if waveform:
            var_names.append('newSpikesNegative')

        translate = {'cluster_id': 'assignedNegative',
                     'timestamp': 'newTimestampsNegative',
                     'waveform': 'newSpikesNegative'}

        # make sure path is a pathlib object
        path = Path(path) if not isinstance(path, Path) else path

        # check if threshold and alignment can be read from the path
        try:
            thresh_str = path.stem
            thresh = float(thresh_str)
        except:
            thresh = np.nan

        try:
            algn = path.parent.stem
            assert algn in ['min', 'max', 'mixed']
        except:
            algn = 'unknown'

    iter_over = tqdm(files) if progress else files
    for fname in iter_over:
        file_path = op.join(path, fname)
        data = loadmat(file_path, squeeze_me=False, variable_names=var_names)
        this_cluster_id = data[translate['cluster_id']].astype('int64')

        if format == 'mm' and isinstance(correct_field, list):
            # TEMP FIX: older exporting function had a spelling error
            correct_field = [field for field in correct_field
                             if field in data][0]
            var_names = read_vars + [correct_field]

        if format == 'mm':
            cluster_id.append(this_cluster_id)
            channel.append([x[0][0] for x in data['channel']])
            alignment.append([x[0][0] for x in data[correct_field]])
            threshold.append(data['threshold'].astype('float32'))

            # unpack to list (and then use .extend)
            timestamp.extend([x[0][0] for x in data['timestamp']])

            # trim first X waveform timesamples (interpolation artifact) ?
            if waveform:
                waveforms.extend([x[0] for x in data['waveform']])
        else:
            # find cluster ids
            cluster_ids = np.unique(this_cluster_id)
            msk = np.in1d(cluster_ids, ignore_cluster)
            if msk.any():
                cluster_ids = cluster_ids[~msk]
            cluster_id.append(cluster_ids)

            # find channel
            ch_name = fname.split('_')[0]
            ch_name = ch_name.replace('CSC', '')

            # split timestamps (and waveforms) into units
            this_timestamps = data[translate['timestamp']]
            for clst_id in cluster_ids:
                unit_msk = this_cluster_id == clst_id
                timestamp.append(this_timestamps[unit_msk])

                alignment.append(algn)
                threshold.append(thresh)
                channel.append(ch_name)

                if waveform:
                    # unit_msk is 1 x n_spikes so here we ignore the
                    # singleton dimension
                    waveforms.append(data[translate['waveform']][unit_msk[0]])

    cluster_id = np.concatenate(cluster_id)
    if format == 'mm':
        alignment = np.concatenate(alignment)
        threshold = np.concatenate(threshold)
        channel = np.concatenate(channel)

        cluster_id = cluster_id[:, 0]
        threshold = threshold[:, 0]

    cellinfo = pd.DataFrame(dict(channel=channel,
                                 cluster=cluster_id,
                                 alignment=alignment,
                                 threshold=threshold))

    # the sfreq here refers to timestamp frequency, not the sampling frequency
    # of the signal
    if waveform:
        n_samples = waveforms[0].shape[1]
        samples_per_ms = 100
        waveform_time = np.arange(n_samples) / samples_per_ms
    else:
        waveform_time = None

    return Spikes(timestamp, sfreq=1e6, cellinfo=cellinfo, waveform=waveforms,
                  waveform_time=waveform_time)


def read_events_neuralynx(path, events_file='Events.nev', format='dataframe',
                          first_timestamp_from='CSC130.ncs'):
    '''Read neuralynx events file as a simple array or dataframe.

    Parameters
    ----------
    path : str | pathlib.Path
        Path to the folder containing neuralynx files.
    events_file : str
        Name of the file containing events. ``'Events.nev'`` by default.
    format : "dataframe" | "mne"
        How to format the events:
        * ``"dataframe"`` - dataframe in BIDS events format
        * ``"mne"`` - array in mne-python format
        defaults to ``"dataframe"``.
    first_timestamp_from : str
        Name of the file to take first timestamp from. ``'CSC130.ncs'`` by
        default. Not used when ``False`` or ``format`` is ``"mne"``.

    Returns
    -------
    events : pandas.DataFrame | numpy.ndarray
        If ``format ="dataframe"``: dataframe containing event times (wrt the
        recording start), trigger values and timestamps. If ``format="mne"``:
        n_events by 3 array in mne-python format (first column - timestamps,
        last columns - trigger values).
    '''
    from .neuralynx_io import load_nev, load_ncs

    events_path = op.join(path, events_file)
    nev = load_nev(events_path)

    # take all trigger timestamps
    event_timestamps = nev['events']['TimeStamp'].astype('int64')

    # take only non-zero event triggers
    triggers = nev['events']['ttl']
    nonzero = triggers > 0
    event_timestamps = event_timestamps[nonzero]
    triggers = triggers[nonzero]

    n_events = event_timestamps.shape[0]
    if format == 'dataframe':
        # take first timestamp of the recording from one of the files
        if first_timestamp_from:
            ncs_path = op.join(path, first_timestamp_from)
            ncs = load_ncs(ncs_path)
            first_sample = ncs['time'][0].astype('int64')
            last_sample = ncs['time'][-1].astype('int64')
            del ncs

        # prepare dataframe
        columns = ['start', 'duration', 'type', 'trigger', 'timestamp']
        if not first_timestamp_from:
            columns.pop(2)

        indices = np.arange(0, n_events + (1 if first_timestamp_from else 0))
        events = pd.DataFrame(columns=columns, index=indices)

        # first row is a special case - info about first timestamp
        # of the recording
        if first_timestamp_from:
            events.loc[0, 'start'] = 0.
            events.loc[0, 'duration'] = 'n/a'
            events.loc[0, 'type'] = 'start'
            events.loc[0, 'trigger'] = -1
            events.loc[0, 'timestamp'] = first_sample
            start, end = 1, n_events
        else:
            start, end = 0, n_events - 1

        # the rest is just copying data to the dataframe
        starts = (event_timestamps - first_sample) / 1e6
        events.loc[start:end, 'start'] = starts
        events.loc[start:end, 'duration'] = 'n/a'
        events.loc[start:end, 'trigger'] = triggers
        events.loc[start:end, 'timestamp'] = event_timestamps

        if first_timestamp_from:
            events.loc[start:end, 'type'] = 'trigger'

            # last timestamp
            events.loc[end + 1, 'start'] = (last_sample - first_sample) / 1e6
            events.loc[end + 1, 'duration'] = 'n/a'
            events.loc[end + 1, 'type'] = 'end'
            events.loc[end + 1, 'trigger'] = -1
            events.loc[end + 1, 'timestamp'] = last_sample

        events= events.infer_objects()

    elif format == 'mne':
        events = np.zeros((n_events, 3), dtype='int64')
        events[:, 0] = event_timestamps
        events[:, -1] = triggers
    else:
        raise ValueError(f'Unknown format "{format}".')

    return events


def read_events_neo(reader, format='mne'):
    '''Read events from a neo reader connected to a file.

    Parameters
    ----------
    reader : neo.io.BaseIO
        Neo reader connected to a file.
    format : "mne"
        How to format the events:
        * ``"mne"`` - array in mne-python format

        At the moment only ``"mne"`` format is supported.

    Returns
    -------
    events : numpy.ndarray
        n_events by 3 array in mne-python format (first column - timestamps,
        last columns - trigger values).
    '''
    assert format == 'mne', 'Only "mne" format is supported at the moment.'
    event_chan = reader.header['event_channels']
    n_event_ch = len(event_chan)

    events = [reader.get_event_timestamps(event_channel_index=idx)
              for idx in range(n_event_ch)]

    # for mne format, we only save events with event_id:
    event_times = list()
    event_ids = list()

    for timestamp, _, event_id in events:
        event_unique = np.unique(event_id)
        if len(event_unique) > 0 and not (event_unique == '').all():
            event_times.append(np.asarray(timestamp))
            event_id = np.asarray(event_id).astype(int)
            event_ids.append(event_id)

    n_elements = len(event_times)
    if n_elements == 0:
        # no events to return - rasie error
        raise RuntimeError('No events in the file.')
    elif n_elements == 1:
        event_times = event_times[0]
        event_ids = event_ids[0]
    else:
        event_times = np.concatenate(event_times)
        event_ids = np.concatenate(event_ids)

        # sort by trigger time
        idx = np.argsort(event_times)
        event_times = event_times[idx]
        event_ids = event_ids[idx]

    # construct MNE-Python triggers array:
    n_events = len(event_times)
    events = np.zeros((n_events, 3), dtype=event_times.dtype)
    events[:, 0] = event_times
    events[:, -1] = event_ids

    return events


def read_events_plexon_nex(path):
    import neo
    reader = neo.io.NeuroExplorerIO(filename=path)
    return read_events_neo(reader, format='mne')


def read_spikes_neo(reader, waveform=True):
    spike_info = reader.header['spike_channels']
    n_units = spike_info.shape[0]
    cell_info = pd.DataFrame(spike_info)

    timestamp_frate = cell_info.wf_sampling_rate.values.copy()
    tstmp_uni = np.unique(timestamp_frate[timestamp_frate > 0])
    assert len(tstmp_uni) == 1
    timestamp_frate = tstmp_uni[0]

    from warnings import warn

    waveform = True

    min_spikes = 10
    use_cells = np.zeros(n_units, dtype=bool)
    waveforms = list() if waveform else None
    timestamps = list()


    for ch_idx in range(n_units):
        tmstmp = reader.get_spike_timestamps(spike_channel_index=ch_idx)
        if min_spikes > 0:
            n_spk = len(tmstmp)
            if n_spk < min_spikes:
                continue

        timestamps.append(np.asarray(tmstmp))
        use_cells[ch_idx] = True

        if waveform:
            wvfm = reader.get_spike_raw_waveforms(spike_channel_index=ch_idx)
            has_wvfm = wvfm is not None

            if has_wvfm:
                wvfm_size = wvfm.shape
                if len(wvfm_size) == 3:
                    # middle dim is the lead dimension
                    n_leads = wvfm_size[1]
                    if n_leads > 1:
                        # warn that we take only the first one
                        warn('Currently if there are multiple leads only the '
                             'waveforms corresponding to the first one are '
                             'read.')
                    wvfm = wvfm.squeeze(axis=1)
                else:
                    wvfm = np.asarray(wvfm)

                # add to waveforms list:
                waveforms.append(wvfm * cell_info.loc[ch_idx, 'wf_gain'])
            else:
                waveforms.append(wvfm)

    cell_info = cell_info.loc[use_cells, :]
    cell_names = cell_info.name.values.copy()

    if waveform:
        waveform_samples = np.array([x.shape[-1] for x in waveforms
                                    if x is not None])
        assert (waveform_samples[0] == waveform_samples).all()
        waveform_samples = waveform_samples[0]
        waveform_time = np.arange(waveform_samples) / (timestamp_frate / 1_000)
    else:
        waveform_time = None

    spk = pln.spikes.Spikes(
        timestamps, sfreq=timestamp_frate, cell_names=cell_names,
        cellinfo=cell_info, waveform=waveforms, waveform_time=waveform_time
    )
    return spk


# TODO: waveforms seem to be read incorrectly for fieldtrip
# sample plexon nex
def read_plexon_nex(path):
    import neo
    reader = neo.io.NeuroExplorerIO(filename=path)
    return read_spikes_neo(reader, format='mne')


# TODO - make sure we can save and write more cellinfo columns
#        (and unit names)
def _convert_spk_to_mm_matlab_format(spk):
    n_units = len(spk)

    data = dict()
    has_cellinfo = spk.cellinfo is not None

    if has_cellinfo:
        columns = spk.cellinfo.columns
        for fld in ['cluster_id', 'alignment', 'channel', 'threshold']:
            col_name = fld.split('_')[0]
            if col_name in columns:
                # extract values from dataframe column
                # and make sure it is N x 1
                data[fld] = spk.cellinfo.loc[:, col_name].values[:, None]

        # other cellinfo columns
        # TODO

    data['timestamp'] = np.empty((n_units, 1), dtype='object')
    data['waveform'] = np.empty((n_units, 1), dtype='object')

    for idx in range(n_units):
        data['timestamp'][idx, 0] = spk.timestamps[idx]
        data['waveform'][idx, 0] = spk.waveform[idx]

    return data


def _save_spk_to_mm_matlab_format(spk, path):
    from scipy.io import savemat

    data = _convert_spk_to_mm_matlab_format(spk)
    savemat(path, data)


def _get_chan_num(chan):
    from numbers import Integral
    if isinstance(chan, Integral):
        return chan
    else:
        return int(''.join([char for char in chan if char.isdigit()]))


def add_region_from_channel_ranges(spk, channel_info, source_column='area',
                                   target_column='region',
                                   channel_column='channel'):
    '''Add brain region information to Spikes from channel ranges info table.

    Parameters
    ----------
    spk : Spikes | SpikeEpochs
        Spikes object whose cell metadata (``.cellinfo``) will be filled
        with information from the table. The ``.cellinfo`` is expected to
        contain channel information in column ``'channel'``. The channel
        information should be either an integer or string, but for strings
        non-numeric values are removed and the remaining numeric values are
        concatenated and turned into integer.
    channel_info : pandas.DataFrame
        Dataframe containing brain region info for specified channel ranges.
        The expected format of this table is the following:
        * ``'channel start'`` column specifying starting values of a channel
          range (for example microwires belonging to the same Behnke-Fried
          electrode).
        * ``'channel end'`` column specifying ending values of a channel range.

        The column specifyng region for given channel range chan be set using
        ``source_column`` argument.
    '''
    assert isinstance(spk.cellinfo, pd.DataFrame)
    assert 'channel' in spk.cellinfo.columns
    chans = spk.cellinfo.channel.unique()

    numeric_rows = [isinstance(x, (int, float))
                    for x in channel_info['channel start']]
    channel_info = channel_info.loc[numeric_rows, :]

    for chan in chans:
        chan_num = _get_chan_num(chan)
        msk = (channel_info['channel start'] <= chan_num) & (
            channel_info['channel end'] >= chan_num)
        region = channel_info.loc[msk, source_column].values[0]

        cell_msk = spk.cellinfo.channel == chan
        spk.cellinfo.loc[cell_msk, target_column] = region



import numpy as np
import matplotlib.pyplot as plt


# TODO - title is now removed, so for groupby it would be good to specify the
#        groupby coord name in legend "title"
# TODO - allow for colors (use ``mpl.colors.to_rgb('C1')`` etc.)
# TODO - the info about "one other dimension" (that is reduced) seems to be no
#        longer accurate

def plot_shaded(arr, reduce_dim=None, groupby=None, ax=None,
                x_dim=None, legend=True, legend_pos=None, colors=None,
                labels=True):
    '''Plot spike rate with standard error of the mean.

    Parameters
    ----------
    frate : xarray.DataArray
        Xarray with at least two dimensions: one is plotted along the x axis
        (this is controlled with ``x_dim`` argument); the other is reduced
        by averaging (see ``reduce_dim`` argument below). The averaged
        dimension also gives rise to the standard error of the mean, which is
        plotted as a shaded area.
    reduce_dim : str
        The dimension to reduce (average). The standard error is also computed
        along this dimension. The default is ``'trial'``.
    groupby : str | None
        The dimension (or sub-dimension) to use as grouping variable plotting
        the spike rate into separate lines. The default is ``None``, which
        does not perform grouping.
    ax : matplotlib.Axes | None
        Axis to plot into. The default is ``None`` which creates an new axis.
    x_dim : str
        Dimension to use for the x axis. The default is ``'time'``.
    legend : bool
        Whether to plot the legend.
    legend_pos : str | None
        Legend position (standard matplotlib names like "upper left"). Defaults
        to ``None`` which uses ``'best'`` position.
    colors : list of arrays | dictionary of arrays | None
        List of RGB arrays to use as colors for condition groups. Can also be
        a dictionary linking condition names / values and RBG arrays. Default
        is ``None`` which uses the default matplotlib color cycle.
    labels : bool
        Whether to add labels to the axes.

    Returns
    -------
    ax : matplotlib.Axes
        Axis with the plot.
    '''
    # auto-infer reduce_dim
    # ---------------------
    if reduce_dim is None:
        auto_reduce_dims = ['trial', 'fold', 'perm', 'permutation', 'cell']
        for dimname in auto_reduce_dims:
            if dimname in arr.coords:
                reduce_dim = dimname
                break

    if ('cell' in arr.coords and not reduce_dim == 'cell'
        and len(arr.cell.shape) > 0):
        if len(arr.coords['cell'] == 1):
            arr = arr.isel(cell=0)
        elif reduce_dim is None:
            reduce_dim = 'cell'
        else:
            msg = ('DataArray contains more than one cell to plot - this is '
                    'not supported.')
            raise RuntimeError(msg)

    # if reduce_dim is still None - use the first dim for 2d array
    # TODO

    # auto-infer x_dim
    # ----------------
    # try time, freq, frequency, lag by default
    if x_dim is None:
        auto_x_dims = ['time', 'frequency', 'freq', 'lag']
        for dimname in auto_x_dims:
            if dimname in arr.coords:
                x_dim = dimname
                break

    # if reduce_dim is still None - use the last dim for 2d array
    # TODO

    ax = plot_xarray_shaded(
        arr, reduce_dim=reduce_dim, x_dim=x_dim, groupby=groupby, ax=ax,
        legend=legend, legend_pos=legend_pos, colors=colors
    )

    if labels:
        xlabel = x_dim.capitalize()
        if 'coord_units' in arr.attrs:
            if x_dim in arr.attrs['coord_units']:
                this_unit = arr.attrs['coord_units'][x_dim]
                xlabel += f' ({this_unit})'
        ax.set_xlabel(xlabel, fontsize=14)

        if arr.name is not None:
            ylabel = arr.name.capitalize()
            if 'unit' in arr.attrs:
                this_unit = arr.attrs['unit']
                ylabel += f' ({this_unit})'
        ax.set_ylabel(ylabel, fontsize=14)

    return ax


# TODO: allow different error shades
def plot_xarray_shaded(arr, reduce_dim=None, x_dim='time', groupby=None,
                       ax=None, legend=True, legend_pos=None, colors=None):
    """
    arr : xarray.DataArray
        Xarray with at least two dimensions: one is plotted along the x axis
        (this is controlled with ``x_dim`` argument); the other is reduced
        by averaging (see ``reduce_dim`` argument below). The averaged
        dimension also gives rise to the standard error of the mean, which is
        plotted as a shaded area.
    reduce_dim : str
        The dimension to reduce (average). The standard error is also computed
        along this dimension. The default is ``'trial'``.
    x_dim : str
        Dimension to use for the x axis. The default is ``'time'``.
    groupby : str | None
        The dimension (or sub-dimension) to use as grouping variable plotting
        the spike rate into separate lines. The default is ``None``, which
        does not perform grouping.
    ax : matplotlib.Axes | None
        Axis to plot into. The default is ``None`` which creates an new axis.
    legend : bool
        Whether to plot the legend.
    legend_pos : str | None
        Legend position (standard matplotlib names like "upper left"). Defaults
        to ``None`` which uses ``'best'`` position.
    colors : list of arrays | dictionary of arrays | None
        List of RGB arrays to use as colors for condition groups. Can also be
        a dictionary linking condition names / values and RBG arrays. Default
        is ``None`` which uses the default matplotlib color cycle.
    """
    assert reduce_dim is not None

    if ax is None:
        _, ax = plt.subplots()

    # compute mean, std and n
    if groupby is not None:
        arr = arr.groupby(groupby)

    # calculate standard error of the mean
    avg = arr.mean(dim=reduce_dim)
    std = arr.std(dim=reduce_dim)
    n = arr.count(dim=reduce_dim)
    std_err = std / np.sqrt(n)
    ci_low = avg - std_err
    ci_high = avg + std_err

    # handle colors
    if colors is not None:
        if groupby is not None:
            group_names = avg.coords[groupby].values
            n_groups = len(group_names)
        else:
            n_groups = 1
            group_names = ['base']

        assert len(colors) == n_groups
        if isinstance(colors, (list, tuple, np.ndarray)):
            assert all(isinstance(x, (list, np.ndarray)) for x in colors)
            colors = {group: color for group, color in zip(group_names, colors)}
        else:
            assert all(name in colors.keys() for name in group_names)
            assert all(isinstance(x, (list, tuple, np.ndarray))
                       for x in colors.values())

    # plot each line with error interval
    if groupby is not None:
        sel = {groupby: 0}
        for val in avg.coords[groupby]:
            val = val.item()
            sel[groupby] = val

            add_arg = {'color': colors[val]} if colors is not None else dict()
            lines = avg.sel(**sel).plot(label=val, ax=ax, **add_arg)
            ax.fill_between(avg.coords[x_dim], ci_low.sel(**sel),
                            ci_high.sel(**sel), alpha=0.3, linewidth=0,
                            **add_arg)
    else:
        add_arg = {'color': colors['base']} if colors is not None else dict()
        lines = avg.plot(ax=ax, **add_arg)
        ax.fill_between(avg.coords[x_dim], ci_low, ci_high, linewidth=0,
                        alpha=0.3, **add_arg)

    if groupby is not None and legend:
        pos = 'best' if legend_pos is None else legend_pos
        ax.legend(title=f'{groupby}:', loc=pos)

    return lines[0].axes


# TODO: move to sarna sometime
# TODO: make sure the pbar is cleared ... (tqdm._instances.clear() may help)
def check_modify_progressbar(pbar, total=None):
    '''Reset ``pbar`` and change its total if it is a tqdm progressbar.
    Otherwise create new progressbar.

    Parameters
    ----------
    pbar : bool | str | tqdm progressbar
        Progressbar to modify or instructions on whether (if bool) or of what
        kind (if str) a progressbar should be created.
    total : int | None
        Total number of steps in the progressbar.

    Returns
    -------
    pbar : tqdm progressbar | empty sarna progressbar
        Modified or newly created progressbar. Can also be an empty progressbar
        from sarna - a progressbar that has some of the tqdm's API but does not
        do anything.
    '''
    if isinstance(pbar, bool):
        if pbar:
            from tqdm import tqdm
            pbar = tqdm(total=total)
        else:
            from sarna.utils import EmptyProgressbar
            pbar = EmptyProgressbar(total=total)
    elif isinstance(pbar, str):
        if pbar == 'notebook':
            from tqdm.notebook import tqdm
        elif pbar == 'text':
            from tqdm import tqdm
        pbar = tqdm(total=total)
    else:
        from tqdm.notebook import tqdm_notebook
        if isinstance(pbar, tqdm_notebook):
            pbar.reset(total=total)
    return pbar


# TODO:
# - [ ] kind='line' ?
# - [ ] datashader backend?
# - [ ] allow to plot multiple average waveforms as lines
def plot_waveform(spk, picks=None, upsample=False, ax=None, labels=True,
                  y_bins=100, times=None):
    '''Plot waveform heatmap for one cell.

    Parameters
    ----------
    spk : pylabianca.spikes.Spikes | pylabianca.spikes.SpikeEpochs
        Spike object to use.
    pick : int | str
        Cell index to plot waveform for.
    upsample : bool | float
        Whether to upsample the waveform (defaults to ``False``). If
        ``True`` the waveform is upsampled by a factor of three. Can also
        be a value to specify the upsampling factor.
    ax : matplotlib.Axes | None
        Axis to plot to. By default opens a new figure.
    labels : bool
        Whether to add labels to the axes.
    y_bins : int
        How many bins to use for the y axis.

    Returns
    -------
    ax : matplotlib.Axes
        Axis with the waveform plot.
    '''
    from .utils import _deal_with_picks

    picks = _deal_with_picks(spk, picks)
    n_picks = len(picks)
    ax = auto_multipanel(n_picks, ax=ax)
    use_ax = _simplify_axes(ax)

    time_unit = 'samples' if times is None else 'ms'

    for idx, unit_idx in enumerate(picks):
        hist, _, ybins, time_edges = _calculate_waveform_density_image(
            spk, unit_idx, upsample, y_bins, times=times
        )
        max_alpha = np.percentile(hist[hist > 0], 45)
        max_lim = np.percentile(hist[hist > 0], 99)

        alpha2 = hist.T * (hist.T <= max_alpha) / max_alpha
        alpha_sum = (hist.T > max_alpha).astype('float') + alpha2
        alpha_sum[alpha_sum > 1] = 1

        use_ax[idx].imshow(
            hist.T, alpha=alpha_sum, vmax=max_lim, origin='lower',
            extent=(time_edges[0], time_edges[-1], ybins[0], ybins[-1]),
            aspect='auto'
        )
        if labels:
            use_ax[idx].set_xlabel(f'Time ({time_unit})')
            use_ax[idx].set_ylabel('Amplitude ($\mu$V)')
            use_ax[idx].set_title(spk.cell_names[unit_idx])

    return ax


def _calculate_waveform_density_image(spk, pick, upsample, y_bins,
                                      density=True, y_range=None, times=None):
    '''Helps in calculating 2d density histogram of the waveforms.'''
    from .utils import _deal_with_picks

    pick = _deal_with_picks(spk, pick)[0]
    n_spikes, n_samples = spk.waveform[pick].shape
    waveform = spk.waveform[pick]
    if upsample:
        if isinstance(upsample, bool):
            upsample = 3
        from scipy import interpolate

        x = np.arange(n_samples)
        interp = interpolate.interp1d(x, waveform, kind='cubic',
                                      assume_sorted=True)

        new_x = np.linspace(0, n_samples - 1, num=n_samples * upsample)
        waveform = interp(new_x)
        n_samples = len(new_x)
    else:
        upsample = 1

    x_coords = np.tile(np.arange(n_samples), (n_spikes, 1))

    # sample_time = 1000 / sfreq  (combinato)
    # sample_time = 1 if times is None else np.diff(times).mean()
    # sample_edge = -94  # -19 for combinato
    # time_edges = [sample_edge * sample_time,
    #               (n_samples / upsample + (sample_edge - 1)) * sample_time]

    if times is not None:
        time_edges = [times[0], times[-1]]
    else:
        time_edges = [0, n_samples / upsample]

    xs = x_coords.ravel()
    ys = waveform.ravel()
    nan_mask = np.isnan(ys)
    if nan_mask.any():
        range = [[np.nanmin(xs), np.nanmax(xs)],
                 [np.nanmin(ys), np.nanmax(ys)]]
        xs = xs[~nan_mask]
        ys = ys[~nan_mask]
    else:
        range = None

    if y_range is not None:
        if nan_mask.any():
            range[1] = y_range
        else:
            range = [[np.min(xs), np.max(xs)], y_range]

    hist, xbins, ybins = np.histogram2d(xs, ys, bins=[n_samples, y_bins],
                                        range=range, density=density)

    return hist, xbins, ybins, time_edges


# TODO: add order=False for groupby?
def plot_raster(spk, pick=0, groupby=None, ax=None, labels=True):
    '''Show spike rasterplot.

    Parameters
    ----------
    spk : pylabianca.spikes.Spikes | pylabianca.spikes.SpikeEpochs
        Spike object to use.
    pick : int | str
        Cell index or name to plot raster for.
    groupby : str | None
        If not None, group the raster by given variable (requires present
        ``.metadata`` field of the ``spk``).
    ax : matplotlib.Axes | None
        Matplotlib axis to plot to. If ``None`` a new figure is opened.
    labels : bool
        Whether to add labels to the axes.

    Returns
    -------
    ax : matplotlib.Axes
        Axis with the raster plot.
    '''

    if ax is None:
        _, ax = plt.subplots()

    spk_cell = spk.copy().pick_cells(picks=pick)

    tri_spikes = list()
    colors = list()

    if groupby is not None:
        values = np.unique(spk_cell.metadata.loc[:, groupby])
    else:
        values = [None]

    for idx, value in enumerate(values):
        img_color = f'C{idx}'
        if groupby is not None:
            trials = (spk_cell.metadata.query(f'{groupby} == {value}')
                      .index.values)
        else:
            if spk_cell.metadata is not None:
                trials = spk_cell.metadata.index.values
            else:
                trials = np.arange(spk_cell.n_trials)

        for trial in trials:
            msk = spk_cell.trial[0] == trial
            tri_spikes.append(spk_cell.time[0][msk])
            colors.append(img_color)

    ax.eventplot(tri_spikes, colors=colors)

    # set y limits
    n_trials = len(tri_spikes)
    ax.set_ylim(-1, n_trials)

    # set x limits
    xlim = spk.time_limits + np.array([-0.05, 0.05])
    ax.set_xlim(xlim)

    if labels:
        ax.set_xlabel('Time (s)', fontsize=14)
        ax.set_ylabel('Trial', fontsize=14)

    return ax


def plot_spikes(spk, frate, groupby=None, df_clst=None, clusters=None,
                pvals=None, pick=0, p_threshold=0.05, min_pval=0.001, ax=None):
    '''Plot average spike rate and spike raster.

    spk : pylabianca.spikes.SpikeEpochs
        Spike object to use.
    frate : xarray.DataArray
        Firing rate xarray in the format returned by ``spk.spike_rate()`` or
        ``spk.spike_density()``.
    groupby : str | None
        How to group the plots. If None, no grouping is done.
    df_clst : pandas.DataFrame | None
        DataFrame with cluster time ranges and p values. If None, no cluster
        information is shown. This argument is to support results obtained
        with ``pylabianca.selectivity.cluster_based_selectivity()``, but one
        can also use the more conventional ``clusters``, ``pvals`` and
        ``p_threshold``.
    clusters : list of np.array
        List of boolean arrays, where each array contains cluster membership
        information (which points along the last array dimension contribute
        to the given cluster).
    pvals : list-like
        List or array of cluster p values.
    p_threshold : float
        Alpha significance threshold. Clusters with p value below this
        threshold will be shown.
    pick : int | str
        Name or index of the cell to plot.
    min_pval : float
        Minimum p-value of cluster to mark on the plot. Only used if
        ``df_clst`` is not None.
    ax: matplotlib.Axes | None
        Two axes to plot on: first is used for average firing ratem the second
        is used for raster plot. If None, a new figure is created.

    Returns
    -------
    fig : matplotlib.Figure
        Figure with the plots.
    '''
    # select cell from frate
    if isinstance(pick, str):
        cell_name = pick
        this_frate = frate.sel(cell=cell_name)
    else:
        this_frate = frate.isel(cell=pick)
        cell_name = this_frate.coords['cell'].item()

    # plot
    if ax is None:
        gridspec_kw = {'bottom': 0.15, 'left': 0.15}
        fig, ax = plt.subplots(nrows=2, gridspec_kw=gridspec_kw)
    else:
        assert(len(ax) == 2)
        fig = ax[0].figure
    plot_shaded(this_frate, groupby=groupby, ax=ax[0])

    # add highlight
    add_highlight = (df_clst is not None) or (
        clusters is not None and pvals is not None)
    if add_highlight:
        if df_clst is not None:
            this_clst = df_clst.query(f'neuron == "{cell_name}"')
            pvals = this_clst.pval.values
            clusters = [_create_mask_from_window_str(twin, this_frate)
                        for twin in this_clst.window.values]
        add_highlights(this_frate, clusters, pvals, ax=ax[0],
                       p_threshold=p_threshold, min_pval=min_pval)

    plot_raster(spk.copy().pick_cells(cell_name), pick=0,
                groupby=groupby, ax=ax[1])
    ylim = ax[1].get_xlim()
    ax[0].set_xlim(ylim)

    ax[0].set_xlabel('')
    ax[0].set_ylabel('Spike rate (Hz)', fontsize=12)
    ax[1].set_xlabel('Time (s)', fontsize=12)
    ax[1].set_ylabel('Trials', fontsize=12)

    return fig


def _create_mask_from_window_str(window, frate):
    twin = [float(x) for x in window.split(' - ')]
    mask = (frate.time.values >= twin[0]) & (frate.time.values <= twin[1])
    return mask


# consider moving to sarna?
# TODO: could infer x coords from plot (if lines are already present)
def add_highlights(arr, clusters, pvals, p_threshold=0.05, ax=None,
                   min_pval=0.001, bottom_extend=True, pval_text=True,
                   text_props=None):
    '''Highlight significant clusters along the last array dimension.

    Parameters
    ----------
    arr : xarray.DataArray | numpy.ndarray
        xarray Data array or numpy array. Used only for axis coordinates.
        If xarray, the last dimension is taken as x axis coordinates.
        If numpy array, the values are assumed to be the x axis coordinates.
    clusters : list of np.array
        List of boolean arrays, where each array contains cluster membership
        information (which points along the last array dimension contribute
        to the given cluster).
    pvals : list-like
        List or array of cluster p values.
    p_threshold : float
        Alpha significance threshold. Clusters with p value below this
        threshold will be shown.
    ax : matplotlib.Axes | None
        Axis to plot to. Optional, defaults to ``None`` - which creates the
        axis automatically.
    min_pval : float
        Minimum meaningful p value. P values below this value will be displayed
        as this value. This is to avoid p = 0, which can result in the cluster
        based permutation test, when no cluster from the null distribution had
        stronger summary statistic than the observed cluster. ``min_pval`` is
        best defined as ``1 / n_permutations``.
    bottom_extend : bool
        Whether to extend the lower limits of y axis when adding bottom
        significance bars. Defaults to ``True``.
    pval_text : bool
        Whether to add p value text boxes to respective cluster ranges in the
        plot. Defaults to ``True``.
    text_props : dict | None
        Dictionary with text properties for p value text boxes. If None,
        defaults to
        ``{'boxstyle': 'round', 'facecolor': 'white', 'alpha': 0.75, edgecolor='gray'}``.

    Returns
    -------
    ax : matplotlib.Axes
        Axis with the plot.
    '''
    from borsar.viz import highlight
    try:
        import xarray as xr
        has_xarray = True
    except ModuleNotFoundError:
        has_xarray = False

    if text_props is None and pval_text:
        text_props = dict(boxstyle='round', facecolor='white', alpha=0.75,
                          edgecolor='gray')

    if ax is None:
        ax = plt.gca()

    if pvals is None:
        return

    pvals_significant = pvals < p_threshold

    if isinstance(arr, np.ndarray):
        x_coords = arr
    elif has_xarray and isinstance(arr, xr.DataArray):
        last_dim = arr.dims[-1]
        x_coords = arr.coords[last_dim].values
    else:
        raise RuntimeError('The `arr` has to be either a numpy array or'
                           'xarray.DataArray.')

    extend_textbox_x = 4

    clusters_x_sorting = np.argsort([np.where(x)[0][0] for x in clusters])

    if pvals_significant.any():
        import borsar

        ylm = ax.get_ylim()
        y_rng = np.diff(ylm)[0]

        # TODO: better estimate text y pos from textbox?
        y_step = 0.075 * y_rng
        text_y = ylm[1] - y_step
        ax.set_ylim([ylm[0], ylm[1] + 2 * y_step])

        sig_idx = np.where(pvals_significant)[0]
        sig_clusters = [clusters[ix] for ix in sig_idx]

        borsar.viz.highlight(
            x_coords, sig_clusters, ax=ax,
            bottom_bar=True, bottom_extend=bottom_extend
        )

        texts = list()
        for ix in clusters_x_sorting:
            if not pvals_significant[ix]:
                continue

            this_pval = pvals[ix]
            text_x = np.mean(x_coords[clusters[ix]])

            if this_pval < min_pval:
                p_txt = 'p < {:.3f}'.format(min_pval)
            else:
                p_txt = borsar.stats.format_pvalue(this_pval)

            this_text = ax.text(
                text_x, text_y, p_txt,
                bbox=text_props, horizontalalignment='center'
            )
            try:
                textbox = this_text.get_window_extent()
            except RuntimeError:
                ax.figure.canvas.draw()
                textbox = this_text.get_window_extent()

            textbox.set_points(
                np.array([
                    [textbox.x0 - extend_textbox_x, textbox.y0],
                    [textbox.x1 + extend_textbox_x, textbox.y1]
                ])
            )

            if len(texts) > 0:
                while textbox.count_overlaps(texts):
                    x, y = this_text.get_position()
                    this_text.set_position((x, y - y_step))
                    textbox = this_text.get_window_extent()
            texts.append(textbox)

    return ax


# - [ ] combine with layout functions from sarna
def align_axes_limits(axes=None, ylim=True, xlim=False):
    '''Align the limits of all ``axes``.'''
    if axes is None:
        axes = plt.gcf().get_axes()

    do_lim = dict(x=xlim, y=ylim)
    limits = dict(x=[np.inf, -np.inf], y=[np.inf, -np.inf])
    iter = (axes if isinstance(axes, (list, np.ndarray))
            else axes.values() if isinstance(axes, dict) else None)

    for ax in iter:
        get_lim = dict(x=ax.get_xlim, y=ax.get_ylim)
        for lim in ('x', 'y'):
            if do_lim[lim]:
                this_lim = get_lim[lim]()
                if limits[lim][0] > this_lim[0]:
                    limits[lim][0] = this_lim[0]
                if limits[lim][1] < this_lim[1]:
                    limits[lim][1] = this_lim[1]

    iter = (axes if isinstance(axes, (list, np.ndarray))
            else axes.values() if isinstance(axes, dict) else None)
    for ax in iter:
        set_lim = dict(x=ax.set_xlim, y=ax.set_ylim)
        for lim in ('x', 'y'):
            if do_lim[lim]:
                set_lim[lim](limits[lim])


# TODO - move this to separate submodule .waveform ?
def calculate_perceptual_waveform_density(spk, cell_idx):
    # get waveform 2d histogram image
    hist, xbins, ybins, time_edges = (
        _calculate_waveform_density_image(
            spk, cell_idx, False, 100)
    )

    # correct y range
    # TODO: would be cheaper to calculate the range in the function above
    sm = hist.sum(axis=0)
    msk = sm < 1e-6
    n_smp = len(msk)

    start_ix, end_ix = 0, n_smp

    if msk[0]:
        # trim from front
        for ix in range(n_smp):
            if not msk[ix]:
                break
        start_ix = ix

    if msk[-1]:
        # trim from back
        for ix in range(-1, -n_smp, -1):
            if not msk[ix]:
                break
        end_ix = ix

    # create the 2d hist image with corrected y range
    y_range = [ybins[start_ix], ybins[end_ix]]
    hist, xbins, ybins, time_edges = (
        _calculate_waveform_density_image(
            spk, cell_idx, False, 100, y_range=y_range)
    )

    # calculate dns
    # TODO: a better way would be to calculate mean from a spatial window
    #       around max, not the top N values
    hist_sort = np.sort(hist)[:, ::-1]
    vals = (hist_sort[:, :15] / hist_sort.max()).mean(axis=-1)
    dns = vals.mean()

    return dns


def auto_multipanel(n_to_show, ax=None, figsize=None):
    '''Create a multipanel figure that fits at least ``n_to_show`` axes.'''
    n = np.sqrt(n_to_show)
    n_left = 0

    if ax is None:
        n *= 1.25
        n_cols = int(np.round(n))
        n_rows = int(np.ceil(n_to_show / n))

        if n_to_show > 1:
            n_left = (n_cols * n_rows) - n_to_show

            # check if one less row works better
            n_cols_try, n_rows_try = n_cols + 1, n_rows - 1
            n_left_try1 = (n_cols_try * n_rows_try) - n_to_show
            try1_good = n_left_try1 > 0 and n_left_try1 < n_left
            n_left_try1 += (1 - try1_good) * 100

            # also check if one less column is better
            n_cols_try2, n_rows_try2 = n_cols - 1, n_rows + 1
            n_left_try2 = (n_cols_try2 * n_rows_try2) - n_to_show
            try2_good = n_left_try2 > 0 and n_left_try2 < n_left
            n_left_try2 += (1 - try2_good) * 100

            if try1_good or try2_good:
                if n_left_try2 < n_left_try1:
                    n_cols, n_rows = n_cols_try2, n_rows_try2
                else:
                    n_cols, n_rows = n_cols_try, n_rows_try

            n_left = (n_cols * n_rows) - n_to_show
            if figsize is None:
                if n_cols == 1 and n_rows == 1:
                    figsize = None
                elif n_rows > 2 or n_cols > 2:
                    # some calculation
                    figsize = (n_cols * 1.35 * 1.5, n_rows * 1.5)

        fig, ax = plt.subplots(nrows=n_rows, ncols=n_cols, figsize=figsize,
                               constrained_layout=True)

        if n_left > 0:
            for this_ax in ax.ravel()[-n_left:]:
                this_ax.axis('off')

    return ax


def _simplify_axes(ax):
    if isinstance(ax, list):
        axes = np.array(ax)
    elif isinstance(ax, np.ndarray):
        axes = ax.ravel()
    else:
        axes = np.array([ax])
    return axes


def plot_isi(spk, picks=None, unit='ms', bins=None, min_spikes=100,
             max_isi=None, ax=None):
    '''Plot inter-spike intervals (ISIs).

    Parameters
    ----------
    spk : pylabianca.spikes.Spikes
        Spikes object to use.
    picks : int | str | list of int | list of str | None
        Which cells to plot. If ``None`` all cells are plotted.
    unit : str
        Time unit to use when plotting the ISIs. Can be ``'ms'`` or ``'s'``.
    bins : int | None
        Number of bins to use for the histograms. If ``None`` the number of
        bins is automatically determined.
    min_spikes : int
        Minimum number of spikes required to plot the ISI histogram.
    max_isi : float | None
        Maximum ISI time to plot. If ``None`` the maximum ISI is set to 0.1 for
        ``unit == 's'`` and 100 for ``unit == 'ms'``.
    ax : matplotlib.Axes | None
        Axis to plot to. If ``None`` a new figure is created.

    Returns
    -------
    ax : matplotlib.Axes
        Axes with the plot.
    '''
    from .utils import _deal_with_picks
    from .spikes import Spikes

    msg = 'Currently only Spikes are supported in ``plot_isi()``.'
    assert(isinstance(spk, Spikes)), msg

    assert unit in ['s', 'ms']
    picks = _deal_with_picks(spk, picks)
    n_picks = len(picks)
    ax = auto_multipanel(n_picks)

    div = 1000 if unit == 'ms' else 1
    if max_isi is None:
        max_isi = 100 if unit == 'ms' else 0.1

    axes = _simplify_axes(ax)
    n_axes = len(axes)
    for idx, unit_idx in enumerate(picks):
        stamps = spk.timestamps[unit_idx]

        if len(stamps) > min_spikes:
            isi = np.diff(stamps / (spk.sfreq / div))
            isi = isi[isi < max_isi]
            n_isi = len(isi)
            use_bins = (bins if bins is not None
                        else min(250, int(n_isi / 100)))
            axes[idx].hist(isi, bins=use_bins)
            axes[idx].set_ylabel('Count', fontsize=12)
            axes[idx].set_xlabel(f'ISI ({unit})', fontsize=12)
        axes[idx].set_title(spk.cell_names[unit_idx], fontsize=12)

    if n_axes > n_picks:
        for ix in range(n_picks, n_axes):
            axes[ix].set_xticks([])
            axes[ix].set_yticks([])

    return ax



# -*- coding: utf-8 -*-
"""
Check region changes after localization.

Created on Tue Jun 20 20:01:40 2023

@author: Asus
"""

import os
import os.path as op

import numpy as np
import pandas as pd


data_dir = r'C:\Users\Asus\OneDrive - Nencki PAN\RESEARCH\anat\derivatives\labels'
files = [f for f in os.listdir(data_dir) if f.endswith('.tsv')]
print(files)


# %%


# %% compare along closest_anat
# SMA -> ACC, ACC -> SMA, SMA -- SMA, ACC -- ACC, + SMA, + ACC

region_transl = {'ACC': 'Anterior-Cingulate', 'SMA': 'Superior-Frontal'}

orig_reg = list(region_transl.keys())
new_reg = list(region_transl.values())
inverse_translate = {val: key for key, val in region_transl.items()}

def classify_row(row):
    case = ''
    this_label = row.label[1:]
    in_orig = this_label in orig_reg
    comp_new = [nw in row.closest_anat for nw in new_reg]
    in_new = any(comp_new)
    
    if not in_new and not in_orig:
        return None

    if in_new:
        matching = new_reg[np.where(comp_new)[0][0]]
        new_old_label = inverse_translate[matching]
        
        if in_orig:
            rel = '--' if this_label == new_old_label else '->'
            case = f'{this_label} {rel} {new_old_label}'
        else:
            case = f'+ {new_old_label}'
    else:
        if in_orig:
            case = f'- {this_label}'
    
    return case


def add_case_to_dict(case, dct):
    if case in dct:
        dct[case] += 1
    else:
        dct[case] = 1


# %%
per_subject = dict()
total = dict()

for fname in files:
    this_subject = dict()
    sub = fname.split('_')[0]

    df = pd.read_csv(op.join(data_dir, fname), sep='\t')
    n_cols = df.shape[0]
    
    for idx in range(n_cols):
        case = classify_row(df.iloc[idx, :])
        
        if case is not None:
            add_case_to_dict(case, this_subject)
            add_case_to_dict(case, total)
    
    per_subject[sub] = this_subject
    
# %%
regions = dict()
regions_no_side = dict()

for fname in files:
    sub = fname.split('_')[0]
    df = pd.read_csv(op.join(data_dir, fname), sep='\t')
    n_cols = df.shape[0]

    for idx in range(n_cols):
        this_area = df.loc[idx, 'closest_anat']
        this_area_no_side = this_area.replace(
            'Left-', '').replace('Right-', '')
        add_case_to_dict(this_area, regions)
        add_case_to_dict(this_area_no_side, regions_no_side)



# -*- coding: utf-8 -*-
"""
Created on Tue Mar 29 15:21:13 2022

@author: mmagnuski
"""
# %%
import os
import os.path as op

import scipy
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from scipy.stats import rankdata
from tqdm import tqdm
import h5io

import pylabianca as pln
import switchorder as swo

# info
# ----
# This is a modified version of the official spke_postproc_choose_ref script
# the modification is to match the way mmagnuski files are read:
# the data are read from BIDS structure, so only subject name is needed
# channel "packs" are no longer assumed to be 8-channel (this is not true for
# some Wrocław files), they are read from channel description files
# first the spikes from selected osort units, then merge info from accompanying
# text file,

# SETTINGS
# --------

subject = 'sub-U12'
session = 'main2'
sorter, norm = 'osort', False
bids_dir = swo.find_switchorder()

dir_fname = (f'{subject}_ses-{session}_task-switchorder_run-01_'
             f'sorter-{sorter}_norm-{norm}')
data_dir = op.join(bids_dir, 'derivatives', 'sorting', subject, f'ses-{session}',
                   dir_fname)
save_fig_dir = op.join(bids_dir, 'derivatives', 'sel_ref', subject,
                       f'ses-{session}')

# whether to plot figures and where to save them
save_fig = True

# data format - 'standard' or 'mm' (depends on how you exported the curated units)
data_format = 'mm'

# minimum coincidence threshold for coincidence cluster formation
coincidence_threshold = 0.1

# minimum number of different-channel units in a coincidence cluster to
# classify it as a reference cluster
min_ref_channels = 1

# weights
# -------
# weights used in calculating the final score (unit with best score is chosen)
# isi - % inter spike intervals < 3 ms (lower is better)
# nspikes  - number of spikes
# snr - signal to noise ratio (mean / std) at alignment point
# std - standard deviation calculated separately for each waveform sample
#       and then averaged (lower is better)
# dns - average "perceptual" waveform density (for each time sample 20 best
#       pixels from waveform density are chosen and averaged; these values
#       are then averaged across the whole waveform; the values are divided
#       by maximum density to de-bias against waveforms with little spikes
#       and thus make this value closer to the density we see in the figures)
weights = {'isi': 0.15, 'nspikes': 0.35, 'snr': 0.0, 'std': 0.0, 'dns': 0.5}

# alignment sample index - 94 should be true for all our osort files
algn_smpl = 94


# %% MAIN SCRIPT

# TODO - add check that highest channel number is in a pack

if not op.exists(save_fig_dir):
    from pathlib import Path
    parent_path = Path(save_fig_dir).parent
    if not op.exists(parent_path):
        os.mkdir(parent_path)
    os.mkdir(save_fig_dir)


# read the file
print('Reading files - including all waveforms...')
spk = swo.read_spk(subject, waveform=True, norm=norm, ses=session)
print('done.')

# remove units with no spikes
n_spikes = spk.n_spikes()
no_spikes = n_spikes == 0
if (no_spikes).any():
    spk.pick_cells(~no_spikes)
    n_spikes = n_spikes[~no_spikes]


def turn_to_percentiles(arr):
    percentiles = np.vectorize(
        lambda x: scipy.stats.percentileofscore(arr, x))(arr)
    return percentiles / 100.


def plot_scores(spk_sel, score):
    from string import ascii_letters

    n_uni = len(spk_sel)
    letters = ascii_letters[1:n_uni + 1]
    ax = plt.figure(constrained_layout=True, figsize=(12, 6)).subplot_mosaic(
        'a' * n_uni + '\n' + letters,
         gridspec_kw={'height_ratios': [3, 1]}
    )

    # plot scores
    ax['a'].plot(score, marker='o')

    # plot best score in red
    best_idx = np.nanargmax(score)
    ax['a'].plot(best_idx, score[best_idx], marker='o', markersize=18,
                 markerfacecolor='r')

    # add waveforms at the bottom
    letters = list(letters)
    for ix in range(len(spk_sel)):
        spk_sel.plot_waveform(ix, ax=ax[letters[ix]], labels=False)
        ax[letters[ix]].set_title(f'{len(spk_sel.timestamps[ix])} spikes')
        ax[letters[ix]].set(yticks=[], xticks=[])

    ax['a'].set_ylabel('Weighted score', fontsize=14)
    return ax['a'].figure


# check channel packs
# -------------------
channel_info = swo.read_channel_info(subject)

# there may be non-numeric rows sometimes (U13 has "?" in some macro channels)
numeric_rows = [isinstance(x, (int, float))
                for x in channel_info['channel start']]
channel_info = channel_info.loc[numeric_rows, :]

# but we want to drop any NaN 'channel start' rows
na_msk = channel_info['channel start'].isna()
if na_msk.any():
    channel_info = channel_info.loc[~na_msk, :]

channels = spk.cellinfo.channel.unique()
channel_number = np.sort([int(ch[1:]) for ch in channels])

first_channel = channel_number.min()
msk = channel_info['channel start'] <= first_channel
first_channel = channel_info.loc[msk, 'channel start'].max()
first_row_idx = np.where(
    channel_info.loc[:, 'channel start'] == first_channel)[0][0]
channel_info_sel = channel_info.iloc[first_row_idx:, :]

# find which units belong to which pack
packs = list()
for _, row in channel_info_sel.iterrows():
    this_pack = np.arange(row['channel start'], row['channel end'] + 1)
    packs.append(this_pack)

unit_channel = spk.cellinfo.channel.str.slice(1).astype('int')
units_in_pack = [np.where(np.in1d(unit_channel, pack))[0]
                 for pack in packs]

# calculate measures
# ------------------
# isi, fr, snr, std, dns

print('Calculating FR, SNR, ISI, STD and DNS...')

n_cells = len(spk)
spk_epo = spk.to_epochs()

measure_names = ['nspikes', 'snr', 'isi', 'std', 'dns']
measures = {name: np.zeros(n_cells, dtype='float')
            for name in measure_names}
measures['nspikes'] = n_spikes

for ix in tqdm(range(n_cells)):
    # SNR
    waveform_ampl = np.mean(spk.waveform[ix][:, algn_smpl], axis=0)
    waveform_std = np.std(spk.waveform[ix][:, algn_smpl], axis=0)
    measures['snr'][ix] = np.abs(waveform_ampl) / waveform_std

    # ISI
    isi_val = np.diff(spk_epo.time[ix])
    prop_below_3ms = (isi_val < 0.003).mean()
    measures['isi'][ix] = prop_below_3ms

    # STD
    avg_std = np.std(spk.waveform[ix], axis=0).mean()
    measures['std'][ix] = avg_std

    # DNS
    measures['dns'][ix] = pln.viz.calculate_perceptual_waveform_density(
        spk, ix)

measures_prc = {name: turn_to_percentiles(measures[name])
                for name in measure_names}
print('Done.')


# compute coincidence
# -------------------
# or read from disk if already computed

fname = 'coincidence_per_pack.hdf5'
files = os.listdir(data_dir)
has_simil = fname in files

group_idx = 0
df_columns=['pack', 'group', 'subgroup', 'channel', 'cluster', 'nspikes',
            'snr', 'isi', 'std', 'dns', 'drop']

if not has_simil:
    # turn to function
    print('Calculating similarity matrix, this can take a few minutes...')
    similarity_per_pack = list()

    for pack_idx in units_in_pack:
        if len(pack_idx) < 2:
            similarity_per_pack.append(None)
        else:
            this_spk = spk.copy().pick_cells(pack_idx)
            similarity = pln.spike_distance.compute_spike_coincidence_matrix(
                this_spk)
            similarity_per_pack.append(similarity)

    print('done.')

    # save to disk
    h5io.write_hdf5(op.join(data_dir, fname), similarity_per_pack,
                    overwrite=True)
else:
    similarity_per_pack = h5io.read_hdf5(op.join(data_dir, fname))
    # add safety checks


# drop = np.zeros(len(spk), dtype='bool')
# drop_perc = drop.copy()
df_list = list()

reordered_names = ['nspikes', 'snr', 'dns', 'isi', 'std']
weights_sel = np.array(
    [weights[name] for name in reordered_names]
)[None, :]

nodrop_simil = list()
for pack_idx, (pack_units_idx, simil) in enumerate(
        zip(units_in_pack, similarity_per_pack)):

    if simil is None:
        continue

    spk_pack = spk.copy().pick_cells(pack_units_idx)
    suspicious_idx, clusters, counts = (
        pln.spike_distance.find_coincidence_clusters(
            simil,
            threshold=coincidence_threshold
        )
    )

    # detect likely REF clusters and select units
    # -------------------------------------------
    check_clst_idx = np.where(counts >= min_ref_channels)[0]

    for cluster_idx in check_clst_idx:
        print(f'Processing pack {pack_idx}, cluster {cluster_idx}...')  # x
        incluster_cell_indices = suspicious_idx[clusters[cluster_idx]]
        cell_idx = pack_units_idx[incluster_cell_indices]
        channels = spk.cellinfo.loc[cell_idx, 'channel'].unique()
        if len(channels) < min_ref_channels:
            continue

        group_idx += 1

        # get, rank and weight measures
        # -----------------------------
        measures_sel = np.stack(
            [measures[name][cell_idx]
             for name in ['nspikes', 'snr', 'dns']
             ] +
            [(1 - measures_prc[name][cell_idx])
             for name in ['isi', 'std']],
            axis=1
        )
        measures_perc_sel = np.stack(
            [measures_prc[name][cell_idx]
             for name in ['nspikes', 'snr', 'dns']
             ] +
            [(1 - measures_prc[name][cell_idx])
             for name in ['isi', 'std']]
        , axis=1)

        score = (measures_perc_sel * weights_sel).sum(axis=1)
        ranks = np.stack(
            [rankdata(measures_sel[:, ix])
             for ix in range(5)]
        , axis=1)
        score_ranks = (ranks * weights_sel).sum(axis=1)

        # save drop cells info according to percentiles and ranks
        # save_idx = cell_idx.copy()
        # save_idx_perc = np.delete(save_idx, score.argmax())
        # save_idx_ranks = np.delete(save_idx, score_ranks.argmax())
        # drop[save_idx_perc] = True
        # drop_perc[save_idx_ranks] = True

        # fill in the dataframe
        # ---------------------
        this_df = (
            spk.cellinfo.loc[cell_idx, ['channel', 'cluster']]
            .reset_index(drop=True)
        )
        this_df.loc[:, 'group'] = group_idx
        this_df.loc[:, 'subgroup'] = np.nan
        this_df.loc[:, 'pack'] = pack_idx + 1

        # add measures
        for ix, measure_name in enumerate(reordered_names):
            this_df.loc[:, measure_name] = measures_sel[:, ix]

        incl_idx = np.ix_(incluster_cell_indices, incluster_cell_indices)
        incluster_simil = simil[incl_idx]

        # iteratively drop according to score_ranks
        subgroup_id = 1
        drop_msk = np.zeros(incluster_simil.shape[0], dtype='bool')
        left_simil = incluster_simil[~drop_msk][:, ~drop_msk]

        while (left_simil > coincidence_threshold).any():
            this_df.loc[~drop_msk, 'subgroup'] = np.nan
            any_left = this_df.subgroup.isna().any()
            while any_left:
                subgroup_idx = np.where(this_df.subgroup.isna())[0]
                best_idx = score_ranks[subgroup_idx].argmax()
                notdrop_idx = subgroup_idx[best_idx]

                other_subgroup_idx = np.delete(subgroup_idx, best_idx)
                other_related = incluster_simil[other_subgroup_idx, notdrop_idx]
                drop = other_related > coincidence_threshold
                other_related = incluster_simil[notdrop_idx, other_subgroup_idx]
                drop = drop | (other_related > coincidence_threshold)
                drop_idx = other_subgroup_idx[drop]
                this_df.loc[drop_idx, 'subgroup'] = subgroup_id
                this_df.loc[drop_idx, 'drop'] = True

                this_df.loc[notdrop_idx, 'subgroup'] = subgroup_id
                this_df.loc[notdrop_idx, 'drop'] = False

                # select the best unit
                any_left = this_df.subgroup.isna().any()
                subgroup_id += 1

            # aggregate nodrop simil
            drop_msk = this_df.loc[:, 'drop'].values.astype('bool')
            left_simil = incluster_simil[~drop_msk][:, ~drop_msk]
        nodrop_simil.append(left_simil)

        df_list.append(this_df)

        # produce and save plots
        # ----------------------
        if save_fig:
            fname_base = f'pack_{pack_idx:02g}_cluster_{cluster_idx:02g}'
            fname = fname_base + '_01_coincid.png'
            fig = pln.spike_distance.plot_high_similarity_cluster(
                spk_pack, simil, clusters, suspicious_idx,
                cluster_idx=cluster_idx)
            fig.savefig(op.join(save_fig_dir, fname), dpi=300)
            plt.close(fig)

            spk_sel = spk.copy().pick_cells(cell_idx)
            # fname = fname_base + '_02_score_percentiles.png'
            # fig = plot_scores(spk_sel, score)
            # fig.savefig(op.join(save_fig_dir, fname), dpi=300)
            # plt.close(fig)

            fname = fname_base + '_03_score_within_cluster_ranks.png'
            fig = plot_scores(spk_sel, score_ranks)
            fig.savefig(op.join(save_fig_dir, fname), dpi=300)
            plt.close(fig)

            # plot after if more than one subgroup
            if left_simil.shape[0] > 1:
                fname = fname_base + '_02_coincid_after.png'
                these_clusters = [clst.copy() for clst in clusters]
                this_clst_msk = these_clusters[cluster_idx]
                this_clst_idx = np.where(this_clst_msk)[0]
                this_clst_msk[this_clst_idx[drop_msk]] = False

                fig = pln.spike_distance.plot_high_similarity_cluster(
                    spk_pack, simil, these_clusters, suspicious_idx,
                    cluster_idx=cluster_idx)
                fig.savefig(op.join(save_fig_dir, fname), dpi=300)
                plt.close(fig)


    if save_fig:
        ignored_clst_idx = np.where(counts < min_ref_channels)[0]

        if len(ignored_clst_idx) > 0:
            print('Plotting ignored clusters...')
            ignored_dir = op.join(save_fig_dir, 'ignored_clusters')
            if not op.isdir(ignored_dir):
                os.mkdir(ignored_dir)

        for cluster_idx in ignored_clst_idx:
            fname_base = f'pack_{pack_idx:02g}_cluster_{cluster_idx:02g}'
            fname = fname_base + '_01_coincid.png'
            fig = pln.spike_distance.plot_high_similarity_cluster(
                spk_pack, simil, clusters, suspicious_idx,
                cluster_idx=cluster_idx)
            fig.savefig(op.join(save_fig_dir, 'ignored_clusters', fname),
                        dpi=300)
            plt.close(fig)

print('Saving dataframe to figures location...')
df = pd.concat(df_list).reset_index(drop=True)
df = df.loc[:, df_columns]
df.loc[:, 'nspikes'] = df.nspikes.astype('int')
df.to_csv(op.join(save_fig_dir, 'table.tsv'), sep='\t')

print('All done.')

# %% write _drop.txt file
# for switchorder the file is read when using switchorder.read_spk() and unit
# rejections are applied
fname = data_dir + '_drop.txt'

lines_to_write = list()
channels_with_drops = np.unique(df.channel.values)

for chan in channels_with_drops:
    df_sel = df.query(f'channel == "{chan}"')
    msk = df_sel.loc[:, 'drop'].values
    if msk.any():
        clusters = df_sel.loc[msk, 'cluster'].values.tolist()
        line = f'{chan}, {clusters}\n'
        lines_to_write.append(line)

with open(fname, 'w') as file:
    file.writelines(lines_to_write)

# %% draw given cluster as graph
import networkx as nx

pack_idx = 3
cluster_idx = 0
plot_edge_threshold = 0.15 # coincidence_threshold

pack_units_idx = units_in_pack[pack_idx]
simil = similarity_per_pack[pack_idx]

spk_pack = spk.copy().pick_cells(pack_units_idx)
suspicious_idx, clusters, counts = (
    pln.spike_distance.find_coincidence_clusters(
        simil,
        threshold=coincidence_threshold
    )
)

idx = suspicious_idx[clusters[cluster_idx]]
simil_clst = simil[idx][:, idx]

# construct graph
G = nx.DiGraph()

n_units = simil_clst.shape[0]
for idx1 in range(n_units):
    for idx2 in range(n_units):
        fromto = simil_clst[idx1, idx2]
        if fromto > plot_edge_threshold:
            G.add_edges_from([(idx1, idx2)], weight=fromto)


# plot'
pos = nx.spring_layout(G, k=1.5)
weights = [G[u][v]['weight'] for u,v in G.edges()]
nx.draw_networkx(G, pos, arrows=True, width=weights)

# %%
spk = swo.read_spk(subject, waveform=False, norm=norm)